# coding=utf-8
# author=uliontse

"""
Copyright 2020 UlionTse

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

   http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

from mlgb.torch.configs import (
    numpy,
    torch,
    BiGRUModeList,
)
from mlgb.torch.functions import (
    IdentityLayer,
    FlattenLayer,
    BatchNormalizationLayer,
    LayerNormalizationLayer,
    ActivationLayer,
    InitializerLayer,
)
from mlgb.error import MLGBError


__all__ = [
    'TaskLayer',
    'LinearLayer',
    'DeepNeuralNetworkLayer',
    'FeedForwardNetworkLayer',
    'Linear3dParallelLayer',
    'DNN3dParallelLayer',
    'Linear2dParallelLayer',
    'ConvolutionalNeuralNetworkLayer',
    'GatedRecurrentUnitLayer',
    'BiGatedRecurrentUnitLayer',
    'CapsuleNetworkLayer',
]


class LinearLayer(torch.nn.Module):
    # tf: default linear_initializer=None='glorot_uniform'; torch: default linear_initializer=None='he_uniform'.
    def __init__(self, linear_unit=1, linear_if_bias=True, linear_activation=None, device='cpu'):
        super().__init__()
        self.linear_unit = linear_unit
        self.linear_if_bias = linear_if_bias
        self.linear_activation = linear_activation
        if self.linear_activation:
            self.activation_fn = ActivationLayer(
                activation=linear_activation,
                device=device,
            )
        self.built = False
        self.device = torch.device(device=device)
        self.to(device=self.device)

    def build(self, x):
        if not self.built:
            if x.ndim != 2:
                raise MLGBError

            f = x.shape[1]
            self.linear_fn = torch.nn.Linear(
                in_features=f,
                out_features=self.linear_unit if self.linear_unit else f,
                bias=self.linear_if_bias,
                device=self.device,
            )
            self.built = True
        return

    def forward(self, x):
        self.build(x)
        x = x.to(device=self.device) if x.device.type != self.device.type else x

        x = self.linear_fn(x)
        if self.linear_activation:
            x = self.activation_fn(x)
        return x


class TaskLayer(torch.nn.Module):
    def __init__(self, task='multiclass:10', task_multiclass_if_project=False, task_multiclass_if_softmax=False,
                 task_multiclass_temperature_ratio=None,
                 task_linear_if_identity=False, task_linear_if_weighted=False, task_linear_if_bias=True,
                 device='cpu'):
        super().__init__()
        self.task = task
        self.task_name = task.split(':')[0]
        if self.task_name not in ('binary', 'regression', 'multiclass'):
            raise MLGBError
        if task_multiclass_temperature_ratio and not (1e-2 <= task_multiclass_temperature_ratio <= 1.0):
            raise MLGBError

        self.task_linear_if_identity = task_linear_if_identity
        self.task_linear_if_weighted = task_linear_if_weighted
        self.task_multiclass_if_project = task_multiclass_if_project
        self.task_multiclass_if_softmax = task_multiclass_if_softmax
        self.task_multiclass_temperature_ratio = task_multiclass_temperature_ratio
        self.task_activation_dict = {
            'regression': None,
            'binary': 'sigmoid',
            'multiclass': 'softmax',
        }
        self.activation_fn = ActivationLayer(
            activation=self.task_activation_dict[self.task_name],
            device=device,
        )
        self.linear_fn = LinearLayer(
            linear_unit=1 if self.task_name != 'multiclass' else int(task.split(':')[1]),
            linear_if_bias=task_linear_if_bias,
            linear_activation=None,
            device=device,
        )
        self.built = False
        self.device = torch.device(device=device)
        self.to(device=self.device)

    def build(self, x):
        if not self.built:
            if x.ndim != 2:
                raise MLGBError
            if self.task_name == 'multiclass' and not self.task_multiclass_if_project:
                if int(self.task.split(':')[1]) != x.shape[1]:
                    raise MLGBError

            self.built = True
        return

    def forward(self, x):
        self.build(x)
        x = x.to(device=self.device) if x.device.type != self.device.type else x

        if self.task_name != 'multiclass':
            if not self.task_linear_if_identity:
                if self.task_linear_if_weighted:
                    x = self.linear_fn(x)
                else:
                    x = torch.sum(x, dim=1, keepdim=True)  # only apply after product of matching model.
            x = self.activation_fn(x)
        else:
            if self.task_multiclass_if_project:
                x = self.linear_fn(x)

            if self.task_multiclass_temperature_ratio:
                x /= self.task_multiclass_temperature_ratio  # zoom data for better performance in softmax_loss.

            if self.task_multiclass_if_softmax:
                x = torch.softmax(x, dim=1)
        return x


class DeepNeuralNetworkLayer(torch.nn.Module):
    def __init__(self, dnn_hidden_units=(64, 32), dnn_activation='relu', dnn_dropout=0.0,
                 dnn_if_bn=False, dnn_if_ln=False, dnn_if_bias=True,
                 device='cpu'):
        super().__init__()
        self.dnn_hidden_units = dnn_hidden_units
        self.dnn_length = len(self.dnn_hidden_units)
        self.dnn_dropouts = [dnn_dropout] * self.dnn_length if isinstance(dnn_dropout, float) else dnn_dropout
        self.dnn_bns = [dnn_if_bn] * self.dnn_length if isinstance(dnn_if_bn, bool) else dnn_if_bn
        self.dnn_lns = [dnn_if_ln] * self.dnn_length if isinstance(dnn_if_ln, bool) else dnn_if_ln
        self.dnn_bias = [dnn_if_bias] * self.dnn_length if isinstance(dnn_if_bias, bool) else dnn_if_bias
        self.dnn_fn_list = torch.nn.ModuleList([
            torch.nn.ModuleList([
                LinearLayer(
                    linear_unit=self.dnn_hidden_units[i],
                    linear_if_bias=self.dnn_bias[i],
                    linear_activation=None,
                    device=device,
                ),
                BatchNormalizationLayer(axis=1, device=device) if self.dnn_bns[i] else IdentityLayer(device=device),
                LayerNormalizationLayer(axis=1, device=device) if self.dnn_lns[i] else IdentityLayer(device=device),
                torch.nn.Dropout(p=self.dnn_dropouts[i]),
                ActivationLayer(activation=dnn_activation, device=device),
            ])
            for i in range(self.dnn_length)
        ])
        self.built = False
        self.device = torch.device(device=device)
        self.to(device=self.device)

    def build(self, x):
        if not self.built:
            if x.ndim != 2:
                raise MLGBError

            self.built = True
        return

    def forward(self, x):
        self.build(x)
        x = x.to(device=self.device) if x.device.type != self.device.type else x

        for dnn_fn in self.dnn_fn_list:
            for fn in dnn_fn:
                x = fn(x)

        return x


class FeedForwardNetworkLayer(torch.nn.Module):
    def __init__(self, ffn_linear_if_twice=True, ffn_if_bias=True, ffn_activation='relu',
                 ffn_dropout=0.0, ffn_if_bn=False, ffn_if_ln=False, ffn_initializer=None,
                 ffn_last_factor=None, ffn_last_activation=None,
                 device='cpu'):
        super().__init__()
        self.ffn_linear_if_twice = ffn_linear_if_twice
        self.ffn_if_bias = ffn_if_bias
        self.ffn_last_factor = ffn_last_factor
        self.ffn_last_activation = ffn_last_activation

        self.bn_fn = BatchNormalizationLayer(axis=1, device=device) if ffn_if_bn else IdentityLayer(device=device)
        self.ln_fn = LayerNormalizationLayer(axis=1, device=device) if ffn_if_ln else IdentityLayer(device=device)
        self.drop_fn = torch.nn.Dropout(p=ffn_dropout)
        self.activation_fn = ActivationLayer(
            activation=ffn_activation,
            device=device,
        )
        if self.ffn_linear_if_twice and self.ffn_last_activation:
            self.activation_last_fn = ActivationLayer(
                activation=ffn_last_activation,
                device=device,
            )
        self.initializer_fn = InitializerLayer(
            initializer=ffn_initializer,
            activation=None,
        ).get()
        self.initializer_zeros_fn = InitializerLayer(
            initializer='zeros',
            activation=None,
        ).get()
        self.built = False
        self.device = torch.device(device=device)
        self.to(device=self.device)

    def build(self, x):
        if not self.built:
            if x.ndim != 3:
                raise MLGBError

            _, f, e = x.shape
            self.ffn_weight_list = [
                self.initializer_fn(torch.nn.parameter.Parameter(
                    data=torch.empty(
                        size=[f, e],
                        device=self.device,
                    ),
                    requires_grad=True,
                ))
                for _ in range(2 if self.ffn_linear_if_twice else 1)
            ]
            if self.ffn_if_bias:
                self.ffn_bias_list = [
                    self.initializer_zeros_fn(torch.nn.parameter.Parameter(
                        data=torch.empty(
                            size=[f, 1],
                            device=self.device,
                        ),
                        requires_grad=True,
                    ))
                    for _ in range(2 if self.ffn_linear_if_twice else 1)
                ]
            self.built = True
        return

    def forward(self, x):
        self.build(x)
        x = x.to(device=self.device) if x.device.type != self.device.type else x

        x = x * self.ffn_weight_list[0] + (self.ffn_bias_list[0] if self.ffn_if_bias else 0.0)
        x = self.bn_fn(x)
        x = self.ln_fn(x)
        x = self.drop_fn(x)
        x = self.activation_fn(x)

        if self.ffn_linear_if_twice:
            x = x * self.ffn_weight_list[1] + (self.ffn_bias_list[1] if self.ffn_if_bias else 0.0)
            if self.ffn_last_activation:
                x = self.activation_last_fn(x)
            if self.ffn_last_factor:
                x = x * self.ffn_last_factor
        return x


class Linear3dParallelLayer(torch.nn.Module):  # vs Dense or EinsumDense with n-d
    def __init__(self, linear_unit=1, linear_activation=None, linear_dropout=0.0, 
                 linear_if_bn=False, linear_if_ln=False, linear_initializer=None, linear_if_bias=True, 
                 device='cpu'):
        super().__init__()
        self.linear_unit = linear_unit
        self.linear_if_bias = linear_if_bias

        self.bn_fn = BatchNormalizationLayer(axis=1, device=device) if linear_if_bn else IdentityLayer(device=device)
        self.ln_fn = LayerNormalizationLayer(axis=1, device=device) if linear_if_ln else IdentityLayer(device=device)
        self.drop_fn = torch.nn.Dropout(p=linear_dropout)
        self.activation_fn = ActivationLayer(activation=linear_activation, device=device)
        self.initializer_fn = InitializerLayer(
            initializer=linear_initializer,
            activation=linear_activation,
        ).get()
        self.initializer_zeros_fn = InitializerLayer(
            initializer='zeros',
            activation=None,
        ).get()
        self.built = False
        self.device = torch.device(device=device)
        self.to(device=self.device)

    def build(self, x):
        if not self.built:
            if x.ndim != 3:
                raise MLGBError

            _, self.linear_parallel_num, self.inputs_width = x.shape

            self.linear_weight = self.initializer_fn(torch.nn.parameter.Parameter(
                data=torch.empty(
                    size=[self.linear_parallel_num, self.inputs_width, self.linear_unit],
                    device=self.device,
                ),
                requires_grad=True,
            ))
            self.linear_bias = self.initializer_zeros_fn(torch.nn.parameter.Parameter(
                data=torch.empty(
                    size=[self.linear_parallel_num, self.linear_unit],
                    device=self.device,
                ),
                requires_grad=True,
            ))
            self.built = True
        return

    def forward(self, x):
        self.build(x)
        x = x.to(device=self.device) if x.device.type != self.device.type else x

        x = torch.einsum('ijk,jkl->ijl', x, self.linear_weight) + (self.linear_bias if self.linear_if_bias else 0.0)
        x = self.bn_fn(x)
        x = self.ln_fn(x)
        x = self.drop_fn(x)
        x = self.activation_fn(x)  # (batch_size, linear_parallel_num, linear_unit)
        return x


class DNN3dParallelLayer(torch.nn.Module):
    def __init__(self, dnn_hidden_units=(64, 32), dnn_activation='relu', dnn_dropout=0.0,
                 dnn_if_bn=False, dnn_if_ln=False, dnn_initializer=None, dnn_if_bias=True, dnn_if_output2d=False, 
                 device='cpu'):
        super().__init__()
        self.dnn_if_output2d = dnn_if_output2d
        self.dnn_hidden_units = dnn_hidden_units
        self.dnn_length = len(self.dnn_hidden_units)
        self.dnn_dropouts = [dnn_dropout] * self.dnn_length if isinstance(dnn_dropout, float) else dnn_dropout
        self.dnn_bns = [dnn_if_bn] * self.dnn_length if isinstance(dnn_if_bn, bool) else dnn_if_bn
        self.dnn_lns = [dnn_if_ln] * self.dnn_length if isinstance(dnn_if_ln, bool) else dnn_if_ln
        self.dnn_bias = [dnn_if_bias] * self.dnn_length if isinstance(dnn_if_bias, bool) else dnn_if_bias

        self.dnn_fn_list = torch.nn.ModuleList([
            Linear3dParallelLayer(
                linear_activation=dnn_activation,
                linear_if_bias=self.dnn_bias[i],
                linear_unit=self.dnn_hidden_units[i],
                linear_dropout=self.dnn_dropouts[i],
                linear_if_bn=self.dnn_bns[i],
                linear_if_ln=self.dnn_lns[i],
                linear_initializer=dnn_initializer,
                device=device,
            ) for i in range(self.dnn_length)
        ])
        self.flatten_fn = FlattenLayer(device=device)
        self.built = False
        self.device = torch.device(device=device)
        self.to(device=self.device)

    def build(self, x):
        if not self.built:
            if x.ndim != 3:
                raise MLGBError

            self.built = True
        return

    def forward(self, x):
        self.build(x)
        x = x.to(device=self.device) if x.device.type != self.device.type else x

        for dnn_fn in self.dnn_fn_list:
            x = dnn_fn(x)

        if self.dnn_if_output2d:
            x = self.flatten_fn(x)
        return x


class Linear2dParallelLayer(torch.nn.Module):
    def __init__(self, linear_parallel_num=1, linear_activation=None, linear_if_bias=True, linear_initializer=None,
                 device='cpu'):
        super().__init__()
        if linear_parallel_num < 1:
            raise MLGBError

        self.linear_parallel_num = linear_parallel_num
        self.linear_unit = 1

        self.linear3d_parallel_fn = Linear3dParallelLayer(
            linear_unit=self.linear_unit,
            linear_activation=linear_activation,
            linear_if_bias=linear_if_bias,
            linear_dropout=0.0,
            linear_if_bn=False,
            linear_if_ln=False,
            linear_initializer=linear_initializer,
            device=device,
        )
        self.flatten_fn = FlattenLayer()
        self.built = False
        self.device = torch.device(device=device)
        self.to(device=self.device)

    def build(self, x):
        if not self.built:
            if x.ndim != 2:
                raise MLGBError

            self.built = True
        return

    def forward(self, x):
        self.build(x)
        x = x.to(device=self.device) if x.device.type != self.device.type else x

        x = torch.stack([x] * self.linear_parallel_num, dim=1)  # (batch_size, linear_parallel_num, inputs_width)
        x = self.linear3d_parallel_fn(x)  # (batch_size, linear_parallel_num, linear_unit)
        x = self.flatten_fn(x)
        return x


class ConvolutionLayer(torch.nn.Module):
    def __init__(self, cnn_conv_mode='Conv1D', cnn_filter_num=64, cnn_kernel_height=32, cnn_kernel_width=1,
                 cnn_activation='tanh', cnn_if_max_pool=True, cnn_pool_size=2, device='cpu'):
        super().__init__()
        if cnn_conv_mode not in ('Conv1D', 'Conv2D'):
            raise MLGBError

        self.cnn_conv_mode = cnn_conv_mode
        self.cnn_if_max_pool = cnn_if_max_pool
        self.cnn_filter_num = cnn_filter_num
        self.cnn_kernel_height = cnn_kernel_height
        self.cnn_kernel_width = cnn_kernel_width
        self.cnn_pool_size = cnn_pool_size
        self.activation_fn = ActivationLayer(
            activation=cnn_activation,
            device=device,
        )
        self.built = False
        self.device = torch.device(device=device)
        self.to(device=self.device)

    def build(self, x):
        if not self.built:
            if x.ndim not in (3, 4):
                raise MLGBError
            if self.cnn_conv_mode == 'Conv1D' and x.ndim != 3:
                raise MLGBError
            if self.cnn_conv_mode == 'Conv2D' and x.ndim != 4:
                raise MLGBError

            if self.cnn_conv_mode == 'Conv1D':
                self.conv_fn = torch.nn.Conv1d(
                    in_channels=x.shape[1],  # torch: data_format='channels_first'.
                    out_channels=self.cnn_filter_num,
                    kernel_size=(self.cnn_kernel_height,),
                    stride=(1,),
                    bias=True,
                    padding='same',
                    padding_mode='zeros',
                    device=self.device,
                )  # (b, e, f) -> (b, cnn_filter_num, f)
                if self.cnn_if_max_pool:
                    self.max_pool_fn = torch.nn.MaxPool1d(
                        kernel_size=(self.cnn_pool_size,),
                        stride=(self.cnn_pool_size,),
                        padding=0,
                    )  # (b, cnn_filter_num, f) -> (b, cnn_filter_num, f // cnn_pool_size)
            else:
                self.conv_fn = torch.nn.Conv2d(
                    in_channels=x.shape[1],  # torch: data_format='channels_first'.
                    out_channels=self.cnn_filter_num,
                    kernel_size=(self.cnn_kernel_height, self.cnn_kernel_width),
                    stride=(1, 1),
                    bias=True,
                    padding='same',
                    padding_mode='zeros',
                    device=self.device,
                )  # (b, 1, f, e) -> (b, cnn_filter_num, f, e)
                if self.cnn_if_max_pool:
                    self.max_pool_fn = torch.nn.MaxPool2d(
                        kernel_size=(self.cnn_pool_size, 1),
                        stride=(self.cnn_pool_size, 1),
                        padding=0,
                    )  # (b, cnn_filter_num, f, e) -> (b, cnn_filter_num, f // cnn_pool_size, e)
            self.built = True
        return

    def forward(self, x):
        self.build(x)
        x = x.to(device=self.device) if x.device.type != self.device.type else x

        x = self.conv_fn(x)
        x = self.activation_fn(x)
        x = self.max_pool_fn(x) if self.cnn_if_max_pool else x
        return x


class ConvolutionalNeuralNetworkLayer(torch.nn.Module):
    def __init__(self, cnn_conv_mode='Conv1D', cnn_filter_nums=(64, 32), cnn_kernel_heights=(64, 32), cnn_kernel_widths=1,
                 cnn_activation='tanh', cnn_if_max_pool=True, cnn_pool_sizes=2, device='cpu'):
        super().__init__()
        if cnn_conv_mode not in ('Conv1D', 'Conv2D'):
            raise MLGBError

        self.cnn_conv_mode = cnn_conv_mode
        self.cnn_length = len(cnn_filter_nums)
        self.cnn_kernel_heights = [cnn_kernel_heights] * self.cnn_length if isinstance(cnn_kernel_heights, int) else cnn_kernel_heights
        self.cnn_kernel_widths = [cnn_kernel_widths] * self.cnn_length if isinstance(cnn_kernel_widths, int) else cnn_kernel_widths
        self.cnn_pool_sizes = [cnn_pool_sizes] * self.cnn_length if isinstance(cnn_pool_sizes, int) else cnn_pool_sizes

        self.cnn_fn_list = torch.nn.ModuleList([
            ConvolutionLayer(
                cnn_conv_mode=cnn_conv_mode,
                cnn_filter_num=cnn_filter_nums[i],
                cnn_kernel_height=self.cnn_kernel_heights[i],
                cnn_kernel_width=self.cnn_kernel_widths[i],
                cnn_activation=cnn_activation,
                cnn_if_max_pool=cnn_if_max_pool,
                cnn_pool_size=self.cnn_pool_sizes[i],
                device=device,
            )
            for i in range(self.cnn_length)
        ])
        self.built = False
        self.device = torch.device(device=device)
        self.to(device=self.device)

    def build(self, x):
        if not self.built:
            if x.ndim not in (3, 4):
                raise MLGBError
            if self.cnn_conv_mode == 'Conv1D' and x.ndim != 3:
                raise MLGBError
            if self.cnn_conv_mode == 'Conv2D' and x.ndim != 4:
                raise MLGBError

            self.built = True
        return

    def forward(self, x):
        self.build(x)
        x = x.to(device=self.device) if x.device.type != self.device.type else x

        for cnn_fn in self.cnn_fn_list:
            x = cnn_fn(x)
        return x


class BaseGatedRecurrentUnitLayer(torch.nn.Module):
    def __init__(self, gru_hidden_unit=32, gru_dropout=0.0, gru_if_bias=True, device='cpu'):
        super().__init__()
        self.gru_hidden_unit = gru_hidden_unit
        self.gru_dropout = gru_dropout
        self.gru_if_bias = gru_if_bias
        self.built = False
        self.device = torch.device(device=device)
        self.to(device=self.device)

    def build(self, x):
        if not self.built:
            if x.ndim != 3:
                raise MLGBError

            self.base_gru_fn = torch.nn.GRU(
                batch_first=True,
                input_size=x.shape[2],  # embed_dim
                hidden_size=self.gru_hidden_unit,
                dropout=self.gru_dropout,
                bias=self.gru_if_bias,
                bidirectional=False,
                num_layers=1,
                device=self.device,
            )
            self.built = True
        return

    def forward(self, x):
        self.build(x)
        x = x.to(device=self.device) if x.device.type != self.device.type else x

        x = self.base_gru_fn(x)[0]
        return x


class GatedRecurrentUnitLayer(torch.nn.Module):
    def __init__(self, gru_hidden_units=(64, 32), gru_dropout=0.0, gru_if_bias=True, device='cpu'):
        super().__init__()
        self.gru_length = len(gru_hidden_units)
        self.gru_fn_list = torch.nn.ModuleList([
            BaseGatedRecurrentUnitLayer(
                gru_hidden_unit=gru_hidden_units[i],
                gru_dropout=gru_dropout,
                gru_if_bias=gru_if_bias,
                device=device,
            )
            for i in range(self.gru_length)
        ])
        self.built = False
        self.device = torch.device(device=device)
        self.to(device=self.device)

    def build(self, x):
        if not self.built:
            if x.ndim != 3:
                raise MLGBError

            self.built = True
        return

    def forward(self, x):
        self.build(x)
        x = x.to(device=self.device) if x.device.type != self.device.type else x

        for gru_fn in self.gru_fn_list:
            x = gru_fn(x)

        x = x[:, -1, :]
        return x


class BiGatedRecurrentUnitLayer(torch.nn.Module):
    def __init__(self, gru_bi_mode='Frontward', gru_hidden_units=(64, 32), gru_dropout=0.0, gru_if_bias=True, device='cpu'):
        super().__init__()
        if gru_bi_mode not in BiGRUModeList:
            raise MLGBError

        self.gru_bi_mode = gru_bi_mode

        self.gru_fn_list = torch.nn.ModuleList([
            GatedRecurrentUnitLayer(
                gru_hidden_units=gru_hidden_units,
                gru_dropout=gru_dropout,
                gru_if_bias=gru_if_bias,
                device=device,
            )
            for _ in range(1 if self.gru_bi_mode in ('Frontward', 'Backward') else 2)
        ])
        self.built = False
        self.device = torch.device(device=device)
        self.to(device=self.device)

    def build(self, x):
        if not self.built:
            if x.ndim != 3:
                raise MLGBError

            self.built = True
        return

    def forward(self, x):
        self.build(x)
        x = x.to(device=self.device) if x.device.type != self.device.type else x
        x_reverse = torch.flip(x, dims=[1])  # x[:, ::-1, :]

        if self.gru_bi_mode == 'Frontward':
            x = self.gru_fn_list[0](x)
        elif self.gru_bi_mode == 'Backward':
            x = self.gru_fn_list[0](x[:, ::-1, :])
        elif self.gru_bi_mode == 'Frontward+Backward':
            x = self.gru_fn_list[0](x) + self.gru_fn_list[1](x_reverse)
        elif self.gru_bi_mode == 'Frontward-Backward':
            x = self.gru_fn_list[0](x) - self.gru_fn_list[1](x_reverse)
        elif self.gru_bi_mode == 'Frontward*Backward':
            x = self.gru_fn_list[0](x) * self.gru_fn_list[1](x_reverse)
        elif self.gru_bi_mode == 'Frontward,Backward':
            x = torch.stack([self.gru_fn_list[0](x), self.gru_fn_list[1](x_reverse)], dim=1)
        else:
            raise MLGBError

        x = x if self.gru_bi_mode == 'Frontward,Backward' else torch.unsqueeze(x, dim=1)
        return x


class CapsuleNetworkLayer(torch.nn.Module):
    def __init__(self, capsule_num=3, capsule_activation='squash', capsule_l2=0.0, capsule_initializer=None,
                 capsule_interest_num_if_dynamic=False, capsule_input_sequence_pad_mode='pre',
                 capsule_routing_initializer='random_normal', device='cpu'):
        super().__init__()
        if capsule_input_sequence_pad_mode not in ('pre', 'post'):
            raise MLGBError

        self.capsule_num = capsule_num
        self.capsule_l2 = capsule_l2
        self.capsule_interest_num_if_dynamic = capsule_interest_num_if_dynamic
        self.capsule_input_sequence_pad_mode = capsule_input_sequence_pad_mode
        self.capsule_initializer_fn = InitializerLayer(
            initializer=capsule_initializer,
            activation=capsule_activation,
        ).get()
        self.capsule_routing_initializer_fn = InitializerLayer(
            initializer=capsule_routing_initializer,
            activation=None,
        ).get()
        self.activation_fn = ActivationLayer(
            activation=capsule_activation,
            device=device,
        )
        self.mask_fn = MaskLayer(
            att_if_mask=True,
            device=device,
        )
        self.built = False
        self.device = torch.device(device=device)
        self.to(device=self.device)

    def build(self, x):
        if not self.built:
            if x.ndim != 3:
                raise MLGBError

            _, seq_len, embed_dim = x.shape
            if self.capsule_interest_num_if_dynamic:
                seq_len = self.get_dynamic_interest_num(seq_len, embed_dim)
                self.seq_len = seq_len

            self.capsule_bilinear_weight = self.capsule_initializer_fn(torch.nn.parameter.Parameter(
                data=torch.empty(
                    size=[seq_len, embed_dim, embed_dim],
                    device=self.device,
                ),
                requires_grad=True,
            ))
            self.capsule_routing_weight = self.capsule_routing_initializer_fn(torch.nn.parameter.Parameter(
                data=torch.empty(
                    size=[1, seq_len, embed_dim],
                    device=self.device,
                ),
                requires_grad=True,
            ))
            self.built = True
        return

    def forward(self, x):
        self.build(x)
        x = x.to(device=self.device) if x.device.type != self.device.type else x

        if self.capsule_interest_num_if_dynamic:
            x = x[:, -self.seq_len:, :] if self.capsule_input_sequence_pad_mode == 'pre' else x[:, :self.seq_len, :]

        w = self.capsule_routing_weight
        w_b = self.capsule_bilinear_weight
        for i in range(self.capsule_num):
            w = self.mask_fn(w)
            w = torch.softmax(w, dim=1)
            x_h = torch.einsum('bfe,fee->bfe', x, w_b) * w  # high_level_capsule: x_h = w * w_b * x
            x_h = self.activation_fn(x_h)  # squash

            w_i = torch.einsum('bfe,fee->bfe', x_h, w_b) * x  # routing_logit: w_i = x_h * w_b * x
            w_i = torch.reduce_sum(w_i, axis=0, keepdims=True)
            w = w + w_i  # if `w` isn't be updated(not bp), it's like `w` of RNN.
            x = x_h
        return x

    def get_dynamic_interest_num(self, seq_len, embed_dim):
        seq_k = max(1, min(seq_len, int(numpy.log2(embed_dim))))
        return seq_k


















































# coding=utf-8
# author=uliontse

"""
Copyright 2020 UlionTse

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

   http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

from mlgb.torch.configs import (
    torch,
    SampleModeList,
)
from mlgb.torch.functions import (
    FlattenLayer,
)
from mlgb.torch.inputs import InputsLayer
from mlgb.torch.components.linears import DeepNeuralNetworkLayer
from mlgb.error import MLGBError


__all__ = [
    'SampledSoftmaxLossLayer',
    'BaseInputsEmbeddingLayer',
]


class SampledSoftmaxLossLayer(torch.nn.Module):
    def __init__(self, sample_mode='Sample:batch', sample_num=None, sample_item_distribution_list=None,
                 sample_fixed_unigram_frequency_list=None, sample_fixed_unigram_distortion=1.0,
                 device='cpu'):
        """
        Paper Link:
            https://arxiv.org/pdf/1310.4546.pdf,
            https://arxiv.org/pdf/1412.2007.pdf,
            https://www.tensorflow.org/extras/candidate_sampling.pdf
        """
        super().__init__()
        if sample_mode not in SampleModeList:
            raise MLGBError

        self.sample_mode = sample_mode
        self.sample_num = sample_num
        self.sample_item_distribution_list = sample_item_distribution_list
        self.sample_fixed_unigram_frequency_list = sample_fixed_unigram_frequency_list
        self.sample_fixed_unigram_distortion = sample_fixed_unigram_distortion
        self.built = False
        self.device = torch.device(device=device)
        self.to(device=self.device)

    def build(self, y_true, y_pred):
        if not self.built:
            if not (y_true.ndim == y_pred.ndim == 2):
                raise MLGBError
            if y_true.shape != y_pred.shape:
                raise MLGBError
            if y_true.shape[1] <= 2:
                raise MLGBError

            self.built = True
        return

    def forward(self, y_true, y_pred):
        y_true = torch.as_tensor(y_true, dtype=torch.float32, device=self.device)
        y_pred = torch.as_tensor(y_pred, dtype=torch.float32, device=self.device)
        self.build(y_true, y_pred)

        if self.sample_mode != 'Sample:batch':
            y_true_int = y_true.to(dtype=torch.int64)
            item_indices = torch.where(torch.sum(y_true_int, dim=0, keepdim=False) > 0)[0]  #

            if self.sample_item_distribution_list:
                q = torch.as_tensor(self.sample_item_distribution_list, dtype=torch.float32, device=self.device)
                q = torch.unsqueeze(q, dim=0)
            else:
                q = self.get_item_distribution(y_true)
            q = torch.index_select(q, dim=1, index=item_indices)

            y_true = torch.index_select(y_true, dim=1, index=item_indices)
            y_pred = torch.index_select(y_pred, dim=1, index=item_indices)
            y_pred -= torch.log(q)

        loss = torch.nn.functional.cross_entropy(input=y_pred, target=y_true, reduction='mean')
        return loss

    def get_item_distribution(self, y_true, add_bias=1):  # Batch Q like BN, not global Q.
        q_top = torch.sum(y_true, dim=0, keepdim=True)
        q_bottom = torch.sum(y_true, dim=None, keepdim=False)
        q = (q_top + add_bias) / (q_bottom + add_bias)  # add 1 avoid log(0) = -inf.
        return q


class BaseInputsEmbeddingLayer(torch.nn.Module):
    def __init__(self, user_feature_names,
                 user_inputs_if_multivalued=False, user_inputs_if_sequential=False, user_inputs_if_embed_dense=True,
                 embed_dim=32, embed_initializer=None,
                 pool_mv_mode='Attention', pool_mv_axis=2, pool_mv_initializer=None,
                 pool_seq_mode='Attention', pool_seq_axis=2, pool_seq_initializer=None,
                 user_dnn_hidden_units=(64, 32), dnn_activation=None, dnn_dropout=0.0,
                 dnn_if_bn=False, dnn_if_ln=False,
                 tower_embeds_flatten_mode='flatten', tower_embeds_if_l2_norm=True,
                 device='cpu'):
        super().__init__()
        if tower_embeds_flatten_mode not in ('flatten', 'sum'):
            raise MLGBError

        self.user_dnn_hidden_units = user_dnn_hidden_units
        self.tower_embeds_flatten_mode = tower_embeds_flatten_mode
        self.tower_embeds_if_l2_norm = tower_embeds_if_l2_norm

        self.user_input_fn = InputsLayer(
            feature_names=user_feature_names,
            inputs_mode='Inputs:feature',
            inputs_if_multivalued=user_inputs_if_multivalued,
            inputs_if_sequential=user_inputs_if_sequential,
            inputs_if_embed_dense=user_inputs_if_embed_dense,
            outputs_dense_if_add_sparse=True,
            embed_dim=embed_dim,
            embed_2d_dim=None,
            embed_cate_if_output2d=False,  #
            embed_initializer=embed_initializer,
            pool_mv_mode=pool_mv_mode,
            pool_mv_axis=pool_mv_axis,
            pool_mv_initializer=pool_mv_initializer,
            pool_seq_mode=pool_seq_mode,
            pool_seq_axis=pool_seq_axis,
            pool_seq_initializer=pool_seq_initializer,
            device=device,
        )
        if self.user_dnn_hidden_units:
            self.dnn_fn = DeepNeuralNetworkLayer(
                dnn_hidden_units=user_dnn_hidden_units,
                dnn_activation=dnn_activation,
                dnn_dropout=dnn_dropout,
                dnn_if_bn=dnn_if_bn,
                dnn_if_ln=dnn_if_ln,
                device=device,
            )
        self.flatten_fn = FlattenLayer(device=device)
        self.built = False
        self.device = torch.device(device=device)
        self.to(device=self.device)

    def build(self, x):  # OriginInputs
        if not self.built:
            if len(x) not in (2, 3, 4):
                raise MLGBError

            self.built = True
        return

    def forward(self, x):
        self.build(x)
        _, x = self.user_input_fn(x)

        if self.tower_embeds_flatten_mode == 'flatten':
            x = self.flatten_fn(x)
        else:
            x = torch.sum(x, dim=1, keepdim=False)

        if self.user_dnn_hidden_units:
            x = self.dnn_fn(x)

        if self.tower_embeds_if_l2_norm:
            x = torch.nn.functional.normalize(x, p=2, dim=1)
        return x






























# coding=utf-8
# author=uliontse

"""
Copyright 2020 UlionTse

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

   http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""


from mlgb.tf.configs import (
    tf,
    Flatten,
    SeqRecPointwiseModeList,
)
from mlgb.tf.functions import (
    FlattenAxesLayer,
    MaskLayer,
    BiasEncoding,
    SimplePoolingLayer,
)
from mlgb.tf.components.linears import (
    DeepNeuralNetworkLayer,
    ConvolutionalNeuralNetworkLayer,
    GatedRecurrentUnitLayer,
    BiGatedRecurrentUnitLayer,
)
from mlgb.tf.components.trms import (
    LabelAttentionLayer,
    TransformerLayer,
)
from mlgb.tf.components.base import (
    LocalActivationUnitLayer,
)
from mlgb.error import MLGBError


__all__ = [
    'GatedRecurrentUnit4RecommendationLayer',
    'ConvolutionalSequenceEmbeddingRecommendationLayer',
    'SelfAttentiveSequentialRecommendationLayer',
    'BidirectionalEncoderRepresentationTransformer4RecommendationLayer',
    'BehaviorSequenceTransformerLayer',
    'DeepInterestNetworkLayer',
    'DeepInterestEvolutionNetworkLayer',
    'DeepSessionInterestNetworkLayer',
]


class GatedRecurrentUnit4RecommendationLayer(tf.keras.layers.Layer):
    def __init__(self, seq_rec_pointwise_mode='Add',
                 gru_hidden_units=(64, 32), gru_activation='tanh', gru_dropout=0.0, gru_l2=0.0, gru_initializer=None,
                 gru_rct_activation='sigmoid', gru_rct_dropout=0.0, gru_rct_l2=0.0, gru_rct_initializer='orthogonal',
                 gru_reset_after=True, gru_unroll=False,
                 dnn_hidden_units=(64, 32), dnn_activation='relu', dnn_dropout=0.0,
                 dnn_if_bn=False, dnn_if_ln=False, dnn_l2=0.0, dnn_initializer=None,
                 seed=None):
        super().__init__()
        if seq_rec_pointwise_mode not in SeqRecPointwiseModeList:
            raise MLGBError
        
        self.seq_rec_pointwise_mode = seq_rec_pointwise_mode

        self.gru_fn = GatedRecurrentUnitLayer(
            gru_hidden_units=gru_hidden_units,
            gru_activation=gru_activation,
            gru_dropout=gru_dropout,
            gru_l2=gru_l2,
            gru_initializer=gru_initializer,
            gru_rct_activation=gru_rct_activation,
            gru_rct_dropout=gru_rct_dropout,
            gru_rct_l2=gru_rct_l2,
            gru_rct_initializer=gru_rct_initializer,
            gru_if_bias=True,
            gru_reset_after=gru_reset_after,
            gru_unroll=gru_unroll,
            seed=seed,
        )
        if self.seq_rec_pointwise_mode in ('LabelAttention', 'Add&LabelAttention'):
            self.label_attention_fn = LabelAttentionLayer(
                softmax_axis=1,
                softmax_pre_temperature_ratio=1.0,
            )
        self.dnn_fn = DeepNeuralNetworkLayer(
            dnn_hidden_units=dnn_hidden_units,
            dnn_activation=dnn_activation,
            dnn_dropout=dnn_dropout,
            dnn_if_bn=dnn_if_bn,
            dnn_if_ln=dnn_if_ln,
            dnn_l2=dnn_l2,
            dnn_initializer=dnn_initializer,
            seed=seed,
        )
        self.flatten_fn_list = [Flatten() for _ in range(2 if self.seq_rec_pointwise_mode in ('Add', 'Add&LabelAttention') else 1)]

    def build(self, input_shape):
        if len(input_shape) != 3:
            raise MLGBError
        if not (input_shape[0].rank == input_shape[1].rank == input_shape[2].rank == 3):
            raise MLGBError

        self.built = True
        return

    @tf.function
    def call(self, inputs):
        # PointWise(not ListWise) by Add&LabelAttention.
        x_user_fea, x_user_seq, x_item_tgt = inputs

        x_seq = self.gru_fn(x_user_seq)  # (b, u == e)
        if self.seq_rec_pointwise_mode in ('LabelAttention', 'Add&LabelAttention'):
            x_k = x_seq
            x_q = tf.reduce_sum(x_item_tgt, axis=1, keepdims=False)
            x_seq = self.label_attention_fn([x_q, x_k])

        x_fea = self.flatten_fn_list[0](x_user_fea)
        x = tf.concat([x_fea, x_seq], axis=1)
        if self.seq_rec_pointwise_mode in ('Add', 'Add&LabelAttention'):
            x_item = self.flatten_fn_list[1](x_item_tgt)
            x = tf.concat([x, x_item], axis=1)

        x = self.dnn_fn(x)
        return x


class ConvolutionalSequenceEmbeddingRecommendationLayer(tf.keras.layers.Layer):
    def __init__(self, seq_rec_pointwise_mode='Add',
                 cnn_filter_num=64, cnn_kernel_size=4, cnn_pool_size=2,
                 cnn_activation='tanh', cnn_l2=0.0, cnn_initializer=None,
                 dnn_hidden_units=(64, 32), dnn_activation='relu', dnn_dropout=0.0,
                 dnn_if_bn=False, dnn_if_ln=False, dnn_l2=0.0, dnn_initializer=None,
                 seed=None):
        super().__init__()
        if seq_rec_pointwise_mode not in SeqRecPointwiseModeList:
            raise MLGBError

        self.seq_rec_pointwise_mode = seq_rec_pointwise_mode
        self.cnn_filter_num = cnn_filter_num
        self.cnn_kernel_size = cnn_kernel_size
        self.cnn_activation = cnn_activation
        self.cnn_l2 = cnn_l2
        self.cnn_initializer = cnn_initializer
        self.cnn_pool_size = cnn_pool_size
        self.seed = seed

        if self.seq_rec_pointwise_mode in ('LabelAttention', 'Add&LabelAttention'):
            self.label_attention_fn = LabelAttentionLayer(
                softmax_axis=1,
                softmax_pre_temperature_ratio=1.0,
            )
        self.dnn_fn = DeepNeuralNetworkLayer(
            dnn_hidden_units=dnn_hidden_units,
            dnn_activation=dnn_activation,
            dnn_dropout=dnn_dropout,
            dnn_if_bn=dnn_if_bn,
            dnn_if_ln=dnn_if_ln,
            dnn_l2=dnn_l2,
            dnn_initializer=dnn_initializer,
            seed=seed,
        )
        self.flatten_fn_list = [Flatten() for _ in range(3 if self.seq_rec_pointwise_mode in ('Add', 'Add&LabelAttention') else 2)]
        self.flatten_axes_fn_list = [FlattenAxesLayer(axes=(1, 3)) for _ in range(2)]

    def build(self, input_shape):
        if len(input_shape) != 3:
            raise MLGBError
        if not (input_shape[0].rank == input_shape[1].rank == input_shape[2].rank == 3):
            raise MLGBError

        self.embed_dim = input_shape[0][2]

        self.h_cnn_fn = ConvolutionalNeuralNetworkLayer(
            cnn_conv_mode='Conv2D',
            cnn_filter_nums=(self.cnn_filter_num,),
            cnn_kernel_heights=(self.cnn_kernel_size,),
            cnn_kernel_widths=(self.embed_dim,),  #
            cnn_activation=self.cnn_activation,
            cnn_l2=self.cnn_l2,
            cnn_initializer=self.cnn_initializer,
            cnn_if_max_pool=True,
            cnn_pool_sizes=(self.cnn_pool_size,),
            seed=self.seed,
        )
        self.v_cnn_fn = ConvolutionalNeuralNetworkLayer(
            cnn_conv_mode='Conv2D',
            cnn_filter_nums=(self.cnn_filter_num,),
            cnn_kernel_heights=(self.cnn_kernel_size,),
            cnn_kernel_widths=(1,),
            cnn_activation=self.cnn_activation,
            cnn_l2=self.cnn_l2,
            cnn_initializer=self.cnn_initializer,
            cnn_if_max_pool=False,  #
            seed=self.seed,
        )
        self.built = True
        return

    @tf.function
    def call(self, inputs):
        # PointWise(not ListWise) by Add&LabelAttention.
        x_user_fea, x_user_seq, x_item_tgt = inputs

        x_user_seq = tf.expand_dims(x_user_seq, axis=3)
        x_h_seq = self.h_cnn_fn(x_user_seq)  # HorizontalConvolution
        x_h_seq = self.flatten_axes_fn_list[0](x_h_seq)
        x_v_seq = self.v_cnn_fn(x_user_seq)  # VerticalConvolution
        x_v_seq = self.flatten_axes_fn_list[1](x_v_seq)
        x_user_seq = tf.concat([x_h_seq, x_v_seq], axis=1)
        
        if self.seq_rec_pointwise_mode in ('LabelAttention', 'Add&LabelAttention'):
            x_k = tf.reduce_sum(x_user_seq, axis=1, keepdims=False)
            x_q = tf.reduce_sum(x_item_tgt, axis=1, keepdims=False)
            x_user_seq = self.label_attention_fn([x_q, x_k])

        x_fea = self.flatten_fn_list[0](x_user_fea)
        x_seq = self.flatten_fn_list[1](x_user_seq)
        x = tf.concat([x_fea, x_seq], axis=1)
        if self.seq_rec_pointwise_mode in ('Add', 'Add&LabelAttention'):
            x_item = self.flatten_fn_list[2](x_item_tgt)
            x = tf.concat([x, x_item], axis=1)

        x = self.dnn_fn(x)
        return x


class Transformer4RecommendationLayer(tf.keras.layers.Layer):
    def __init__(self, seq_rec_pointwise_mode='Add',
                 trm_num=1, trm_if_pe=True, trm_mha_head_num=1, trm_mha_head_dim=32,
                 trm_mha_if_mask=True, trm_mha_l2=0.0, trm_mha_initializer=None,
                 trm_if_ffn=True, trm_ffn_activation='gelu', trm_ffn_l2=0.0, trm_ffn_initializer=None,
                 trm_residual_dropout=0.0,
                 dnn_hidden_units=(64, 32), dnn_activation='relu', dnn_dropout=0.0,
                 dnn_if_bn=False, dnn_if_ln=False, dnn_l2=0.0, dnn_initializer=None,
                 seed=None):
        super().__init__()
        if seq_rec_pointwise_mode not in SeqRecPointwiseModeList + ['OnlyAddTrm']:
            raise MLGBError

        self.seq_rec_pointwise_mode = seq_rec_pointwise_mode

        self.trm_fn_list = [
            TransformerLayer(
                trm_if_pe=True if (i == 0 and trm_if_pe) else False,
                trm_mha_head_num=trm_mha_head_num,
                trm_mha_head_dim=trm_mha_head_dim,
                trm_mha_if_mask=trm_mha_if_mask,
                trm_mha_l2=trm_mha_l2,
                trm_mha_initializer=trm_mha_initializer,
                trm_if_ffn=trm_if_ffn,
                trm_ffn_activation=trm_ffn_activation,
                trm_ffn_l2=trm_ffn_l2,
                trm_ffn_initializer=trm_ffn_initializer,
                trm_residual_dropout=trm_residual_dropout,
                seed=seed + i if isinstance(seed, int) else seed,
            )
            for i in range(trm_num)
        ]
        if self.seq_rec_pointwise_mode in ('LabelAttention', 'Add&LabelAttention'):
            self.label_attention_fn = LabelAttentionLayer(
                softmax_axis=1,
                softmax_pre_temperature_ratio=1.0,
            )
        self.dnn_fn = DeepNeuralNetworkLayer(
            dnn_hidden_units=dnn_hidden_units,
            dnn_activation=dnn_activation,
            dnn_dropout=dnn_dropout,
            dnn_if_bn=dnn_if_bn,
            dnn_if_ln=dnn_if_ln,
            dnn_l2=dnn_l2,
            dnn_initializer=dnn_initializer,
            seed=seed,
        )
        self.flatten_fn_list = [Flatten() for _ in range(3 if self.seq_rec_pointwise_mode in ('Add', 'Add&LabelAttention') else 2)]

    def build(self, input_shape):
        if len(input_shape) != 3:
            raise MLGBError
        if not (input_shape[0].rank == input_shape[1].rank == input_shape[2].rank == 3):
            raise MLGBError

        self.built = True
        return

    @tf.function
    def call(self, inputs):
        # PointWise(not ListWise) by Add&LabelAttention.
        x_user_fea, x_user_seq, x_item_tgt = inputs

        if self.seq_rec_pointwise_mode == 'OnlyAddTrm':
            x_user_seq = tf.concat([x_user_seq, x_item_tgt], axis=1)
        
        for trm_fn in self.trm_fn_list:
            x_user_seq = trm_fn([x_user_seq, x_user_seq])

        if self.seq_rec_pointwise_mode in ('LabelAttention', 'Add&LabelAttention'):
            x_k = tf.reduce_sum(x_user_seq, axis=1, keepdims=False)
            x_q = tf.reduce_sum(x_item_tgt, axis=1, keepdims=False)
            x_user_seq = self.label_attention_fn([x_q, x_k])

        x_fea = self.flatten_fn_list[0](x_user_fea)
        x_seq = self.flatten_fn_list[1](x_user_seq)
        x = tf.concat([x_fea, x_seq], axis=1)
        if self.seq_rec_pointwise_mode in ('Add', 'Add&LabelAttention'):
            x_item = self.flatten_fn_list[2](x_item_tgt)
            x = tf.concat([x, x_item], axis=1)

        x = self.dnn_fn(x)
        return x
        

class SelfAttentiveSequentialRecommendationLayer(Transformer4RecommendationLayer):
    def __init__(self, seq_rec_pointwise_mode='Add',
                 trm_mha_head_num=8, trm_mha_head_dim=32,
                 trm_mha_if_mask=True, trm_mha_l2=0.0, trm_mha_initializer=None,
                 trm_if_ffn=True, trm_ffn_activation='gelu', trm_ffn_l2=0.0, trm_ffn_initializer=None,
                 trm_residual_dropout=0.0,
                 dnn_hidden_units=(64, 32), dnn_activation='relu', dnn_dropout=0.0,
                 dnn_if_bn=False, dnn_if_ln=False, dnn_l2=0.0, dnn_initializer=None,
                 seed=None):
        super().__init__(
            seq_rec_pointwise_mode=seq_rec_pointwise_mode,
            trm_num=1,
            trm_if_pe=False,
            trm_mha_head_num=trm_mha_head_num,
            trm_mha_head_dim=trm_mha_head_dim,
            trm_mha_if_mask=trm_mha_if_mask,
            trm_mha_l2=trm_mha_l2,
            trm_mha_initializer=trm_mha_initializer,
            trm_if_ffn=trm_if_ffn,
            trm_ffn_activation=trm_ffn_activation,
            trm_ffn_l2=trm_ffn_l2,
            trm_ffn_initializer=trm_ffn_initializer,
            trm_residual_dropout=trm_residual_dropout,
            dnn_hidden_units=dnn_hidden_units,
            dnn_activation=dnn_activation,
            dnn_dropout=dnn_dropout,
            dnn_if_bn=dnn_if_bn,
            dnn_if_ln=dnn_if_ln,
            dnn_l2=dnn_l2,
            dnn_initializer=dnn_initializer,
            seed=seed,
        )


class BidirectionalEncoderRepresentationTransformer4RecommendationLayer(Transformer4RecommendationLayer):
    def __init__(self, seq_rec_pointwise_mode='Add',
                 trm_num=4, trm_if_pe=True, trm_mha_head_num=8, trm_mha_head_dim=32,
                 trm_mha_if_mask=True, trm_mha_l2=0.0, trm_mha_initializer=None,
                 trm_if_ffn=True, trm_ffn_activation='gelu', trm_ffn_l2=0.0, trm_ffn_initializer=None,
                 trm_residual_dropout=0.0,
                 dnn_hidden_units=(64, 32), dnn_activation='relu', dnn_dropout=0.0,
                 dnn_if_bn=False, dnn_if_ln=False, dnn_l2=0.0, dnn_initializer=None,
                 seed=None):
        super().__init__(
            seq_rec_pointwise_mode=seq_rec_pointwise_mode,
            trm_num=trm_num,
            trm_if_pe=trm_if_pe,
            trm_mha_head_num=trm_mha_head_num,
            trm_mha_head_dim=trm_mha_head_dim,
            trm_mha_if_mask=trm_mha_if_mask,
            trm_mha_l2=trm_mha_l2,
            trm_mha_initializer=trm_mha_initializer,
            trm_if_ffn=trm_if_ffn,  # PositionWiseFFN but unnecessary, nonlinearity should rely on activation of FFN.
            trm_ffn_activation=trm_ffn_activation,
            trm_ffn_l2=trm_ffn_l2,
            trm_ffn_initializer=trm_ffn_initializer,
            trm_residual_dropout=trm_residual_dropout,
            dnn_hidden_units=dnn_hidden_units,
            dnn_activation=dnn_activation,
            dnn_dropout=dnn_dropout,
            dnn_if_bn=dnn_if_bn,
            dnn_if_ln=dnn_if_ln,
            dnn_l2=dnn_l2,
            dnn_initializer=dnn_initializer,
            seed=seed,
        )


class BehaviorSequenceTransformerLayer(Transformer4RecommendationLayer):
    def __init__(self, trm_mha_head_num=8, trm_mha_head_dim=32,
                 trm_mha_if_mask=True, trm_mha_l2=0.0, trm_mha_initializer=None,
                 trm_if_ffn=True, trm_ffn_activation='gelu', trm_ffn_l2=0.0, trm_ffn_initializer=None,
                 trm_residual_dropout=0.0,
                 dnn_hidden_units=(64, 32), dnn_activation='selu', dnn_dropout=0.0,
                 dnn_if_bn=False, dnn_if_ln=False, dnn_l2=0.0, dnn_initializer=None,
                 seed=None):
        super().__init__(
            seq_rec_pointwise_mode='OnlyAddTrm',
            trm_num=1,
            trm_if_pe=True,
            trm_mha_head_num=trm_mha_head_num,
            trm_mha_head_dim=trm_mha_head_dim,
            trm_mha_if_mask=trm_mha_if_mask,
            trm_mha_l2=trm_mha_l2,
            trm_mha_initializer=trm_mha_initializer,
            trm_if_ffn=trm_if_ffn,
            trm_ffn_activation=trm_ffn_activation,
            trm_ffn_l2=trm_ffn_l2,
            trm_ffn_initializer=trm_ffn_initializer,
            trm_residual_dropout=trm_residual_dropout,
            dnn_hidden_units=dnn_hidden_units,
            dnn_activation=dnn_activation,
            dnn_dropout=dnn_dropout,
            dnn_if_bn=dnn_if_bn,
            dnn_if_ln=dnn_if_ln,
            dnn_l2=dnn_l2,
            dnn_initializer=dnn_initializer,
            seed=seed,
        )


class DeepInterestNetworkLayer(tf.keras.layers.Layer):
    def __init__(self, lau_version='v4', lau_hidden_units=(16,), lau_if_softmax=False,
                 dnn_hidden_units=(64, 32), dnn_activation='dice', dnn_dropout=0.0, dnn_if_bn=False, dnn_if_ln=False,
                 dnn_l2=0.0, dnn_initializer=None,
                 seed=None):
        super().__init__()
        self.lau_if_softmax = lau_if_softmax

        self.lau_fn = LocalActivationUnitLayer(
            lau_version=lau_version,
            lau_hidden_units=lau_hidden_units,
            lau_activation=dnn_activation,
            seed=seed,
        )
        self.dnn_fn = DeepNeuralNetworkLayer(
            dnn_hidden_units=dnn_hidden_units,
            dnn_activation=dnn_activation,
            dnn_dropout=dnn_dropout,
            dnn_if_bn=dnn_if_bn,
            dnn_if_ln=dnn_if_ln,
            dnn_l2=dnn_l2,
            dnn_initializer=dnn_initializer,
            seed=seed,
        )
        if self.lau_if_softmax:
            self.mask_fn = MaskLayer(att_if_mask=True)
        self.flatten_fn = Flatten()

    def build(self, input_shape):
        if len(input_shape) != 3:
            raise MLGBError
        if not (input_shape[0].rank == input_shape[1].rank == input_shape[2].rank == 3):
            raise MLGBError
        if input_shape[1] != input_shape[2]:
            raise MLGBError

        self.built = True
        return

    @tf.function
    def call(self, inputs):
        x_b, x_q, x_k = inputs  # (base, query, key==value)
        x_v = x_k
        x_b = self.flatten_fn(x_b)

        w = self.lau_fn([x_q, x_k])  # (b, s, 1)
        if self.lau_if_softmax:
            w = self.mask_fn(w)
            w = tf.nn.softmax(w, axis=1)  # axis -> s

        x_v = x_v * w
        x_v = tf.reduce_sum(x_v, axis=2, keepdims=False)
        x = tf.concat([x_v, x_b], axis=1)
        x = self.dnn_fn(x)
        return x


class DeepInterestEvolutionNetworkLayer(GatedRecurrentUnit4RecommendationLayer):
    def __init__(self, gru_hidden_units=(64, 32), gru_activation='tanh', gru_dropout=0.0, gru_l2=0.0, gru_initializer=None,
                 gru_rct_activation='sigmoid', gru_rct_dropout=0.0, gru_rct_l2=0.0, gru_rct_initializer='orthogonal',
                 gru_reset_after=True, gru_unroll=False,
                 dnn_hidden_units=(64, 32), dnn_activation='dice', dnn_dropout=0.0,
                 dnn_if_bn=False, dnn_if_ln=False, dnn_l2=0.0, dnn_initializer=None,
                 seed=None):
        super().__init__(
            seq_rec_pointwise_mode='Add&LabelAttention',  #
            gru_hidden_units=gru_hidden_units,
            gru_activation=gru_activation,
            gru_dropout=gru_dropout,
            gru_l2=gru_l2,
            gru_initializer=gru_initializer,
            gru_rct_activation=gru_rct_activation,
            gru_rct_dropout=gru_rct_dropout,
            gru_rct_l2=gru_rct_l2,
            gru_rct_initializer=gru_rct_initializer,
            gru_reset_after=gru_reset_after,
            gru_unroll=gru_unroll,
            dnn_hidden_units=dnn_hidden_units,
            dnn_activation=dnn_activation,  #
            dnn_dropout=dnn_dropout,
            dnn_if_bn=dnn_if_bn,
            dnn_if_ln=dnn_if_ln,
            dnn_l2=dnn_l2,
            dnn_initializer=dnn_initializer,
            seed=seed,
        )


class DeepSessionInterestNetworkLayer(tf.keras.layers.Layer):
    def __init__(self, dsin_if_process_session=False, session_pool_mode='Pooling:average', session_size=4, session_stride=2,
                 bias_l2=0.0, bias_initializer='zeros',
                 trm_mha_head_num=4, trm_mha_head_dim=32,
                 trm_mha_if_mask=True, trm_mha_l2=0.0, trm_mha_initializer=None,
                 trm_if_ffn=True, trm_ffn_activation='gelu', trm_ffn_l2=0.0, trm_ffn_initializer=None,
                 trm_residual_dropout=0.0,
                 gru_bi_mode='Frontward,Backward',
                 gru_hidden_units=(64, 32), gru_activation='tanh', gru_dropout=0.0, gru_l2=0.0, gru_initializer=None,
                 gru_rct_activation='sigmoid', gru_rct_dropout=0.0, gru_rct_l2=0.0, gru_rct_initializer='orthogonal',
                 gru_reset_after=True, gru_unroll=False,
                 dnn_hidden_units=(64, 32), dnn_activation='relu', dnn_dropout=0.0,
                 dnn_if_bn=False, dnn_if_ln=False, dnn_l2=0.0, dnn_initializer=None,
                 seed=None):
        super().__init__()
        self.dsin_if_process_session = dsin_if_process_session
        self.session_size = session_size
        self.session_stride = session_stride

        if self.dsin_if_process_session:
            self.pool_fn = SimplePoolingLayer(
                pool_mode=session_pool_mode,
                pool_axis=1,
                pool_axis_if_keep=True,  #
            )
        self.bias_fn = BiasEncoding(
            if_bias=True,
            bias_l2=bias_l2,
            bias_initializer=bias_initializer,
            seed=seed,
        )
        self.trm_fn = TransformerLayer(
            trm_if_pe=False,
            trm_mha_head_num=trm_mha_head_num,
            trm_mha_head_dim=trm_mha_head_dim,
            trm_mha_if_mask=trm_mha_if_mask,
            trm_mha_l2=trm_mha_l2,
            trm_mha_initializer=trm_mha_initializer,
            trm_if_ffn=trm_if_ffn,
            trm_ffn_activation=trm_ffn_activation,
            trm_ffn_l2=trm_ffn_l2,
            trm_ffn_initializer=trm_ffn_initializer,
            trm_residual_dropout=trm_residual_dropout,
            seed=seed,
        )
        self.bi_gru_fn = BiGatedRecurrentUnitLayer(
            gru_bi_mode=gru_bi_mode,
            gru_hidden_units=gru_hidden_units,
            gru_activation=gru_activation,
            gru_dropout=gru_dropout,
            gru_l2=gru_l2,
            gru_initializer=gru_initializer,
            gru_rct_activation=gru_rct_activation,
            gru_rct_dropout=gru_rct_dropout,
            gru_rct_l2=gru_rct_l2,
            gru_rct_initializer=gru_rct_initializer,
            gru_if_bias=True,
            gru_reset_after=gru_reset_after,
            gru_unroll=gru_unroll,
            seed=seed,
        )
        self.label_attention_fn_list = [
            LabelAttentionLayer(
                softmax_axis=1,
                softmax_pre_temperature_ratio=1.0,
            )
            for _ in range(2)
        ]
        self.dnn_fn = DeepNeuralNetworkLayer(
            dnn_hidden_units=dnn_hidden_units,
            dnn_activation=dnn_activation,
            dnn_dropout=dnn_dropout,
            dnn_if_bn=dnn_if_bn,
            dnn_if_ln=dnn_if_ln,
            dnn_l2=dnn_l2,
            dnn_initializer=dnn_initializer,
            seed=seed,
        )
        self.flatten_fn_list = [Flatten() for _ in range(2)]

    def build(self, input_shape):
        if len(input_shape) != 3:
            raise MLGBError
        if not (input_shape[0].rank == input_shape[1].rank == input_shape[2].rank == 3):
            raise MLGBError

        if self.dsin_if_process_session:
            seq_length = input_shape[1][1]
            if seq_length < self.session_size + self.session_stride * 2:
                raise MLGBError('seq_length < self.session_size + self.session_stride * 2')

        self.built = True
        return

    @tf.function
    def call(self, inputs):
        x_user_fea, x_user_seq, x_item_tgt = inputs

        x_user_seq = self.seq2session2seq(x_user_seq)
        x_user_seq = self.bias_fn(x_user_seq)

        x_trm_seq = self.trm_fn([x_user_seq, x_user_seq])
        x_gru_seq = self.bi_gru_fn(x_trm_seq)

        x_q = tf.reduce_sum(x_item_tgt, axis=1, keepdims=False)
        x_trm_k = tf.reduce_sum(x_trm_seq, axis=1, keepdims=False)
        x_gru_k = tf.reduce_sum(x_gru_seq, axis=1, keepdims=False)
        x_trm_seq = self.label_attention_fn_list[0]([x_q, x_trm_k])
        x_gru_seq = self.label_attention_fn_list[1]([x_q, x_gru_k])

        x_fea = self.flatten_fn_list[0](x_user_fea)
        x_item = self.flatten_fn_list[1](x_item_tgt)
        x = tf.concat([x_fea, x_item, x_trm_seq, x_gru_seq], axis=1)
        x = self.dnn_fn(x)
        return x

    def seq2session2seq(self, x):
        if self.dsin_if_process_session:
            x_list = [
                self.pool_fn(x[:, i: i+self.session_size, :])
                for i in range(0, x.shape[1] - (self.session_size + self.session_stride), self.session_stride)
            ]
            x = tf.concat(x_list, axis=1)
        return x

































# coding=utf-8
# author=uliontse

"""
Copyright 2020 UlionTse

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

   http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

from mlgb.tf.configs import (
    numpy,
    tf,
    L1L2,
    Dropout,
    Flatten,
    BatchNormalization,
    LayerNormalization,
    FBIModeList,
)
from mlgb.tf.functions import (
    IdentityLayer,
    ActivationLayer,
    InitializerLayer,
    KMaxPoolingLayer,
    TransposeLayer,
    FlattenAxesLayer,
)
from mlgb.tf.components.linears import (
    LinearLayer,
    DeepNeuralNetworkLayer,
    DNN3dParallelLayer,
    Linear2dParallelLayer,
    ConvolutionalNeuralNetworkLayer,
)
from mlgb.tf.components.fbis import (
    AllFieldBinaryInteractionLayer,
    GroupedAllFieldWiseBinaryInteractionLayer,
)
from mlgb.tf.components.trms import (
    InteractingLayer,
)
from mlgb.tf.components.base import (
    MaskBlockLayer,
    ResidualUnitLayer,
    CrossNetworkLayer,
    RegulationModuleLayer,
    BridgeModuleLayer,
    FactorEstimatingNetworkLayer,
    LogarithmicTransformationLayer,
    CompressedInteractionNetworkLayer,
    SqueezeExcitationNetworkLayer,
)
from mlgb.error import MLGBError


__all__ = [
    'LinearOrLogisticRegressionLayer',
    'MultiLayerPerceptronLayer',
    'PiecewiseLinearModelLayer',
    'DeepLearningRecommendationModelLayer',
    'MaskNetLayer',

    'DeepCrossingModelLayer',
    'DeepCrossNetworkLayer',
    'EnhancedDeepCrossNetworkLayer',
    'AllFactorizationMachineLayer',
    'FactorizationMachineLayer',
    'FieldFactorizationMachineLayer',
    'FieldWeightedFactorizationMachineLayer',
    'FieldEmbeddedFactorizationMachineLayer',
    'FieldVectorizedFactorizationMachineLayer',
    'FieldMatrixedFactorizationMachineLayer',
    'AttentionalFactorizationMachineLayer',
    'LorentzFactorizationMachineLayer',
    'InteractionMachineLayer',
    'AllInputFactorizationMachineLayer',
    'InputFactorizationMachineLayer',
    'DualInputFactorizationMachineLayer',

    'FactorizationMachineNeuralNetworkLayer',
    'ProductNeuralNetworkLayer',
    'ProductNetworkInNetworkLayer',
    'OperationNeuralNetworkLayer',
    'AdaptiveFactorizationNetworkLayer',

    'NeuralFactorizationMachineLayer',
    'WideDeepLearningLayer',
    'DeepFactorizationMachineLayer',
    'DeepFieldEmbeddedFactorizationMachineLayer',
    'FieldLeveragedEmbeddingNetworkLayer',

    'ConvolutionalClickPredictionModelLayer',
    'FeatureGenerationByConvolutionalNeuralNetworkLayer',
    'ExtremeDeepFactorizationMachineLayer',
    'FeatureImportanceBilinearInteractionNetworkLayer',
    'AutomaticFeatureInteractionLearningLayer',
]


class LinearOrLogisticRegressionLayer(LinearLayer):
    def __init__(self, linear_if_bias=True, linear_l1=0.0, linear_l2=0.0, linear_initializer=None, seed=None):
        super().__init__(
            linear_if_bias=linear_if_bias,
            linear_l1=linear_l1,
            linear_l2=linear_l2,
            linear_initializer=linear_initializer,
            seed=seed,
        )


class MultiLayerPerceptronLayer(DeepNeuralNetworkLayer):
    def __init__(self, dnn_hidden_units=(64, 32), dnn_activation='tanh', dnn_dropout=0.0,
                 dnn_if_bn=False, dnn_if_ln=False, dnn_if_bias=True, dnn_l2=0.0, dnn_initializer=None, seed=None):
        super().__init__(
            dnn_hidden_units=dnn_hidden_units,
            dnn_activation=dnn_activation,
            dnn_dropout=dnn_dropout,
            dnn_if_bn=dnn_if_bn,
            dnn_if_ln=dnn_if_ln,
            dnn_if_bias=dnn_if_bias,
            dnn_l2=dnn_l2,
            dnn_initializer=dnn_initializer,
            seed=seed,
        )


class PiecewiseLinearModelLayer(tf.keras.layers.Layer):
    def __init__(self, plm_task='binary', plm_piece_num=2,
                 linear_l1=0.0, linear_l2=0.0, linear_initializer=None, seed=None):
        super().__init__()
        self.task_activation_dict = {
            'regression': None,
            'binary': 'sigmoid',
        }
        self.linear2d_parallel_fn = Linear2dParallelLayer(
            linear_parallel_num=plm_piece_num,
            linear_activation=None,
            linear_if_bias=True,
            linear_l1=linear_l1,
            linear_l2=linear_l2,
            linear_initializer=linear_initializer,
            seed=seed,
        )
        self.task2d_parallel_fn = Linear2dParallelLayer(
            linear_parallel_num=plm_piece_num,
            linear_activation=self.task_activation_dict[plm_task],
            linear_if_bias=True,
            linear_l1=linear_l1,
            linear_l2=linear_l2,
            linear_initializer=linear_initializer,
            seed=seed,
        )

    def build(self, input_shape):
        if len(input_shape) != 2:
            raise MLGBError
        if not (input_shape[0].rank == input_shape[1].rank == 2):
            raise MLGBError

        self.built = True
        return

    @tf.function
    def call(self, inputs):
        piece_dense_2d_tensor, base_dense_2d_tensor = inputs

        piece_tensor = self.linear2d_parallel_fn(piece_dense_2d_tensor)  # (batch_size, plm_piece_num)
        base_tensor = self.task2d_parallel_fn(base_dense_2d_tensor)  # (batch_size, plm_piece_num)
        piece_w_tensor = tf.nn.softmax(piece_tensor, axis=1)
        plm_outputs = base_tensor * piece_w_tensor
        plm_outputs = tf.reduce_sum(plm_outputs, axis=1, keepdims=True)
        return plm_outputs


class DeepLearningRecommendationModelLayer(tf.keras.layers.Layer):
    def __init__(self, dnn_bottom_hidden_units=(64, 32), dnn_top_hidden_units=(64, 32),
                 dnn_activation='relu', dnn_dropout=0.0, dnn_if_bn=False, dnn_if_ln=False,
                 dnn_l2=0.0, dnn_initializer=None, seed=None):
        super().__init__()
        self.dnn_bottom_fn = DeepNeuralNetworkLayer(
            dnn_hidden_units=dnn_bottom_hidden_units,
            dnn_activation=dnn_activation,
            dnn_dropout=dnn_dropout,
            dnn_if_bn=dnn_if_bn,
            dnn_if_ln=dnn_if_ln,
            dnn_l2=dnn_l2,
            dnn_initializer=dnn_initializer,
            seed=seed,
        )
        self.dnn_top_fn = DeepNeuralNetworkLayer(
            dnn_hidden_units=dnn_top_hidden_units,
            dnn_activation=dnn_activation,
            dnn_dropout=dnn_dropout,
            dnn_if_bn=dnn_if_bn,
            dnn_if_ln=dnn_if_ln,
            dnn_l2=dnn_l2,
            dnn_initializer=dnn_initializer,
            seed=seed,
        )

    def build(self, input_shape):
        if len(input_shape) != 2:
            raise MLGBError
        if not (input_shape[0].rank == input_shape[1].rank == 2):
            raise MLGBError

        self.built = True
        return

    @tf.function
    def call(self, inputs):
        dense_2d_tensor, embed_2d_tensor = inputs

        bottom_outputs = self.dnn_bottom_fn(dense_2d_tensor)
        top_inputs = tf.concat([bottom_outputs, embed_2d_tensor], axis=1)
        dlrm_outputs = self.dnn_top_fn(top_inputs)
        return dlrm_outputs


class MaskNetLayer(tf.keras.layers.Layer):
    def __init__(self, mask_mode='MaskNet:serial', mask_block_num=4,
                 block_activation='relu', block_if_bn=False, block_dropout=0.0, block_l2=0.0, block_initializer=None,
                 dnn_hidden_units=(64, 32), dnn_activation='relu',
                 dnn_if_bn=False, dnn_if_ln=False, dnn_dropout=0.0, dnn_l2=0.0, dnn_initializer=None,
                 seed=None):
        super().__init__()
        if mask_mode not in ('MaskNet:serial', 'MaskNet:parallel'):
            raise MLGBError
        if not mask_block_num > 0:
            raise MLGBError

        self.mask_mode = mask_mode

        self.block_fn_list = [
            MaskBlockLayer(
                ffn_activation=block_activation,
                ffn_if_bn=block_if_bn,
                ffn_dropout=block_dropout,
                ffn_l2=block_l2,
                ffn_initializer=block_initializer,
                seed=seed + i if isinstance(seed, int) else seed,
            )
            for i in range(mask_block_num)
        ]
        if self.mask_mode == 'MaskNet:parallel':
            self.dnn_fn = DeepNeuralNetworkLayer(
                dnn_hidden_units=dnn_hidden_units,
                dnn_activation=dnn_activation,
                dnn_dropout=dnn_dropout,
                dnn_if_bn=dnn_if_bn,
                dnn_if_ln=dnn_if_ln,
                dnn_l2=dnn_l2,
                dnn_initializer=dnn_initializer,
                seed=seed,
            )
        self.ln_emb_fn = LayerNormalization(axis=2)  #
        self.flatten_fn = Flatten()

    def build(self, input_shape):
        if input_shape.rank != 3:
            raise MLGBError

        self.built = True
        return

    @tf.function
    def call(self, inputs):
        embed_3d_tensor = inputs

        x_e = self.ln_emb_fn(embed_3d_tensor)
        x_m = embed_3d_tensor

        if self.mask_mode == 'MaskNet:parallel':
            x_pool = [block_fn([x_e, x_m]) for block_fn in self.block_fn_list]
            x = tf.concat(x_pool, axis=1)
            x = self.flatten_fn(x)
            x = self.dnn_fn(x)
        else:
            for block_fn in self.block_fn_list:
                x_e = block_fn([x_e, x_m])
            x = self.flatten_fn(x_e)
        return x


class DeepCrossingModelLayer(tf.keras.layers.Layer):
    def __init__(self, dcm_if_dnn=True, dcm_if_ln=False,
                 dnn_hidden_units=(64, 32), dnn_activation='relu',
                 dnn_dropout=0.0, dnn_if_bn=False, dnn_if_ln=False, dnn_l2=0.0, dnn_initializer=None,
                 seed=None):
        super().__init__()
        self.dcm_layer_length = len(dnn_hidden_units)
        self.dcm_fn_list = [
            ResidualUnitLayer(
                dcm_if_dnn=dcm_if_dnn,
                dcm_if_ln=dcm_if_ln,
                dnn_hidden_unit=dnn_hidden_units[i],
                dnn_activation=dnn_activation,
                dnn_dropout=dnn_dropout,
                dnn_if_bn=dnn_if_bn,
                dnn_if_ln=dnn_if_ln,
                dnn_l2=dnn_l2,
                dnn_initializer=dnn_initializer,
                seed=seed + i if isinstance(seed, int) else seed,
            )
            for i in range(self.dcm_layer_length)
        ]

    def build(self, input_shape):
        if input_shape.rank != 2:
            raise MLGBError

        self.built = True
        return

    @tf.function
    def call(self, inputs):
        x = inputs
        for dcm_fn in self.dcm_fn_list:
            x = dcm_fn(x)
        return x


class DeepCrossNetworkLayer(tf.keras.layers.Layer):
    def __init__(self, dcn_version='v2', dcn_l2=0.0, dcn_initializer=None,
                 dnn_hidden_units=(64, 32), dnn_activation='relu', dnn_dropout=0.0, dnn_if_bn=False, dnn_if_ln=False,
                 dnn_l2=0.0, dnn_initializer=None,
                 seed=None):
        super().__init__()
        self.cross_fn = CrossNetworkLayer(
            dcn_layer_num=len(dnn_hidden_units),
            dcn_version=dcn_version,
            dcn_l2=dcn_l2,
            dcn_initializer=dcn_initializer,
            seed=seed,
        )
        self.dnn_fn = DeepNeuralNetworkLayer(
            dnn_hidden_units=dnn_hidden_units,
            dnn_activation=dnn_activation,
            dnn_dropout=dnn_dropout,
            dnn_if_bn=dnn_if_bn,
            dnn_if_ln=dnn_if_ln,
            dnn_l2=dnn_l2,
            dnn_initializer=dnn_initializer,
            seed=seed,
        )

    def build(self, input_shape):
        if input_shape.rank != 2:
            raise MLGBError

        self.built = True
        return

    @tf.function
    def call(self, inputs):
        x = inputs
        cn_outputs = self.cross_fn(x)
        dnn_outputs = self.dnn_fn(x)
        outputs = tf.concat([cn_outputs, dnn_outputs], axis=-1)
        return outputs


class EnhancedDeepCrossNetworkLayer(tf.keras.layers.Layer):
    def __init__(self, edcn_layer_num=2,
                 bdg_mode='EDCN:attention_pooling', bdg_layer_num=1,
                 rgl_tau_ratio=1.0, rgl_initializer='ones', rgl_l2=0.0,
                 dcn_version='v2', dcn_l2=0.0, dcn_initializer=None,
                 dnn_activation='relu', dnn_dropout=0.0, dnn_if_bn=False, dnn_if_ln=False,
                 dnn_l2=0.0, dnn_initializer=None,
                 seed=None):
        super().__init__()
        self.edcn_layer_num = edcn_layer_num
        self.dnn_activation = dnn_activation
        self.dnn_dropout = dnn_dropout
        self.dnn_if_bn = dnn_if_bn
        self.dnn_if_ln = dnn_if_ln
        self.dnn_l2 = dnn_l2
        self.dnn_initializer = dnn_initializer
        self.seed = seed

        self.rgl_cross_fn_list = [
            RegulationModuleLayer(
                fgu_tau_ratio=rgl_tau_ratio,
                fgu_initializer=rgl_initializer,
                fgu_l2=rgl_l2,
                seed=seed + i if isinstance(seed, int) else seed,
            ) for i in range(edcn_layer_num+1)
        ]
        self.rgl_deep_fn_list = [
            RegulationModuleLayer(
                fgu_tau_ratio=rgl_tau_ratio,
                fgu_initializer=rgl_initializer,
                fgu_l2=rgl_l2,
                seed=seed + i if isinstance(seed, int) else seed,
            ) for i in range(edcn_layer_num+1)
        ]
        self.bdg_fn_list = [
            BridgeModuleLayer(
                bdg_mode=bdg_mode,
                bdg_layer_num=bdg_layer_num,
                dnn_activation=dnn_activation,
                dnn_dropout=dnn_dropout,
                dnn_if_bn=dnn_if_bn,
                dnn_if_ln=dnn_if_ln,
                dnn_l2=dnn_l2,
                dnn_initializer=dnn_initializer,
                seed=seed + i if isinstance(seed, int) else seed,
            ) for i in range(edcn_layer_num+1)
        ]
        self.cross_fn_list = [
            CrossNetworkLayer(
                dcn_layer_num=1,
                dcn_version=dcn_version,
                dcn_l2=dcn_l2,
                dcn_initializer=dcn_initializer,
                seed=seed + i if isinstance(seed, int) else seed,
            ) for i in range(edcn_layer_num*2)
        ]

    def build(self, input_shape):
        if input_shape.rank != 3:
            raise MLGBError

        _, self.fields_width, self.embed_dim = input_shape
        self.inputs_width = int(self.fields_width * self.embed_dim)

        self.dnn_fn_list = [
            DeepNeuralNetworkLayer(
                dnn_hidden_units=[self.inputs_width],
                dnn_activation=self.dnn_activation,
                dnn_dropout=self.dnn_dropout,
                dnn_if_bn=self.dnn_if_bn,
                dnn_l2=self.dnn_l2,
                dnn_initializer=self.dnn_initializer,
                seed=self.seed + i if isinstance(self.seed, int) else self.seed,
            ) for i in range(self.edcn_layer_num*2)
        ]
        self.built = True
        return

    @tf.function
    def call(self, inputs):
        x = inputs
        x_c = self.rgl_cross_fn_list[-1](x)  # last
        x_d = self.rgl_deep_fn_list[-1](x)  # last

        for i in range(self.edcn_layer_num):
            x_c = self.cross_fn_list[i*2](x_c)
            x_d = self.dnn_fn_list[i*2](x_d)

            x_bdg = self.bdg_fn_list[i]([x_c, x_d])
            x = tf.reshape(x_bdg, shape=[-1, self.fields_width, self.embed_dim])

            x_c = self.rgl_cross_fn_list[i](x)
            x_d = self.rgl_deep_fn_list[i](x)
            x_c = self.cross_fn_list[i*2+1](x_c)
            x_d = self.dnn_fn_list[i*2+1](x_d)

        x_bdg = self.bdg_fn_list[-1]([x_c, x_d])  # last
        x = tf.concat([x_c, x_d, x_bdg], axis=1)
        return x


class AllFactorizationMachineLayer(tf.keras.layers.Layer):
    def __init__(self, fbi_mode='FM', fbi_unit=32, fbi_l2=0.0, fbi_initializer=None,
                 fbi_hofm_order=3, fbi_afm_activation='relu', fbi_afm_dropout=0.0,
                 linear_l1=0.0, linear_l2=0.0, linear_initializer=None,
                 seed=None):
        super().__init__()
        self.fbi_fn = AllFieldBinaryInteractionLayer(
            fbi_mode=fbi_mode,
            fbi_unit=fbi_unit,
            fbi_l2=fbi_l2,
            fbi_initializer=fbi_initializer,
            fbi_hofm_order=fbi_hofm_order,
            fbi_afm_activation=fbi_afm_activation,
            fbi_afm_dropout=fbi_afm_dropout,
            seed=seed,
        )
        self.linear_fn = LinearLayer(
            linear_initializer=linear_initializer,
            linear_l1=linear_l1,
            linear_l2=linear_l2,
            seed=seed,
        )

    def build(self, input_shape):
        if len(input_shape) != 2:
            raise MLGBError
        if input_shape[0].rank != 2:
            raise MLGBError
        if input_shape[1].rank not in (2, 3):
            raise MLGBError

        self.built = True
        return

    @tf.function
    def call(self, inputs):
        dense_2d_tensor, embed_3d_tensor = inputs

        lr_outputs = self.linear_fn(dense_2d_tensor)
        fbi_outputs = self.fbi_fn(embed_3d_tensor)
        fm_outputs = tf.concat([lr_outputs, fbi_outputs], axis=-1)
        return fm_outputs


class FactorizationMachineLayer(AllFactorizationMachineLayer):
    def __init__(self, fbi_mode='FM', fbi_unit=32, fbi_l2=0.0, fbi_initializer=None,
                 linear_l1=0.0, linear_l2=0.0, linear_initializer=None, seed=None):
        super().__init__(
            fbi_mode, fbi_unit, fbi_l2, fbi_initializer,
            3, 'relu', 0.0,
            linear_l1, linear_l2, linear_initializer, seed,
        )
        if fbi_mode not in ('FM', 'FM3D'):
            raise MLGBError


class FieldFactorizationMachineLayer(AllFactorizationMachineLayer):
    def __init__(self, fbi_mode='FFM', fbi_unit=32, fbi_l2=0.0, fbi_initializer=None,
                 linear_l1=0.0, linear_l2=0.0, linear_initializer=None, seed=None):
        super().__init__(
            fbi_mode, fbi_unit, fbi_l2, fbi_initializer,
            3, 'relu', 0.0,
            linear_l1, linear_l2, linear_initializer, seed,
        )
        if fbi_mode != 'FFM':
            raise MLGBError


class HigherOrderFactorizationMachineLayer(AllFactorizationMachineLayer):
    def __init__(self, fbi_mode='HOFM', fbi_unit=32, fbi_l2=0.0, fbi_initializer=None,
                 fbi_hofm_order=3,
                 linear_l1=0.0, linear_l2=0.0, linear_initializer=None, seed=None):
        super().__init__(
            fbi_mode, fbi_unit, fbi_l2, fbi_initializer,
            fbi_hofm_order, 'relu', 0.0,
            linear_l1, linear_l2, linear_initializer, seed,
        )
        if fbi_mode != 'HOFM':
            raise MLGBError


class FieldWeightedFactorizationMachineLayer(AllFactorizationMachineLayer):
    def __init__(self, fbi_mode='FwFM', fbi_unit=32, fbi_l2=0.0, fbi_initializer=None,
                 linear_l1=0.0, linear_l2=0.0, linear_initializer=None, seed=None):
        super().__init__(
            fbi_mode, fbi_unit, fbi_l2, fbi_initializer,
            3, 'relu', 0.0,
            linear_l1, linear_l2, linear_initializer, seed,
        )
        if fbi_mode != 'FwFM':
            raise MLGBError


class FieldEmbeddedFactorizationMachineLayer(AllFactorizationMachineLayer):
    def __init__(self, fbi_mode='FEFM', fbi_unit=32, fbi_l2=0.0, fbi_initializer=None,
                 linear_l1=0.0, linear_l2=0.0, linear_initializer=None, seed=None):
        super().__init__(
            fbi_mode, fbi_unit, fbi_l2, fbi_initializer,
            3, 'relu', 0.0,
            linear_l1, linear_l2, linear_initializer, seed,
        )
        if fbi_mode != 'FEFM':
            raise MLGBError


class FieldVectorizedFactorizationMachineLayer(AllFactorizationMachineLayer):
    def __init__(self, fbi_mode='FvFM', fbi_unit=32, fbi_l2=0.0, fbi_initializer=None,
                 linear_l1=0.0, linear_l2=0.0, linear_initializer=None, seed=None):
        super().__init__(
            fbi_mode, fbi_unit, fbi_l2, fbi_initializer,
            3, 'relu', 0.0,
            linear_l1, linear_l2, linear_initializer, seed,
        )
        if fbi_mode != 'FvFM':
            raise MLGBError


class FieldMatrixedFactorizationMachineLayer(AllFactorizationMachineLayer):
    def __init__(self, fbi_mode='FmFM', fbi_unit=32, fbi_l2=0.0, fbi_initializer=None,
                 linear_l1=0.0, linear_l2=0.0, linear_initializer=None, seed=None):
        super().__init__(
            fbi_mode, fbi_unit, fbi_l2, fbi_initializer,
            3, 'relu', 0.0,
            linear_l1, linear_l2, linear_initializer, seed,
        )
        if fbi_mode != 'FmFM':
            raise MLGBError


class AttentionalFactorizationMachineLayer(tf.keras.layers.Layer):
    def __init__(self, fbi_mode='AFM', fbi_unit=32, fbi_l2=0.0, fbi_initializer=None,
                 fbi_afm_activation='relu', fbi_afm_dropout=0.0,
                 linear_l1=0.0, linear_l2=0.0, linear_initializer=None,
                 seed=None):
        super().__init__(
            fbi_mode, fbi_unit, fbi_l2, fbi_initializer,
            3, fbi_afm_activation, fbi_afm_dropout,
            linear_l1, linear_l2, linear_initializer, seed,
        )
        if fbi_mode != 'AFM':
            raise MLGBError


class LorentzFactorizationMachineLayer(tf.keras.layers.Layer):
    def __init__(self, lfm_beta=1.0, fbi_l2=0.0, fbi_initializer=None, seed=None):
        super().__init__()
        self.lfm_beta = lfm_beta
        self.fbi_l2 = fbi_l2
        self.fbi_initializer = InitializerLayer(
            initializer=fbi_initializer,
            activation=None,
            seed=seed,
        ).get()

    def build(self, input_shape):
        if input_shape.rank != 3:
            raise MLGBError

        self.fields_width = input_shape[1]
        self.product_width = int(self.fields_width * (self.fields_width-1) // 2)
        self.ij_ids = numpy.array(numpy.triu_indices(n=self.fields_width, k=1)).T

        self.fbi_weight = self.add_weight(
            name='fbi_weight',
            shape=input_shape.as_list()[1:],
            initializer=self.fbi_initializer,
            regularizer=L1L2(0.0, self.fbi_l2),
            trainable=True,
        )
        self.built = True
        return

    @tf.function
    def call(self, inputs):
        x = inputs
        x = x * self.fbi_weight

        u_i = tf.gather(params=x, indices=tf.constant(self.ij_ids[:, 0], dtype=tf.int32), axis=1)
        v_i = tf.gather(params=x, indices=tf.constant(self.ij_ids[:, 1], dtype=tf.int32), axis=1)
        u_0 = self.lorentzian_norm_x(u_i)
        v_0 = self.lorentzian_norm_x(v_i)
        uv_lip = self.lorentzian_inner_product(u_i, v_i, u_0, v_0)

        x = self.triangle_pooling(uv_lip, u_0, v_0)
        return x

    def lorentzian_norm_x(self, x):
        x = tf.square(tf.reduce_sum(tf.pow(x, 2), axis=2, keepdims=False) + self.lfm_beta)
        return x

    def lorentzian_inner_product(self, u_i, v_i, u_0, v_0):
        x = tf.reduce_sum(u_i * v_i, axis=2, keepdims=False) - u_0 * v_0
        return x

    def triangle_pooling(self, uv_lip, u_0, v_0):
        x = (1 - uv_lip - u_0 - v_0) / (u_0 * v_0)  # (batch_size, product_width)
        return x


class InteractionMachineLayer(tf.keras.layers.Layer):
    def __init__(self, im_mode='IM', im_order=3, im_l2=0.0, im_initializer=None,
                 dnn_hidden_units=(64, 32), dnn_activation='relu', dnn_dropout=0.0, dnn_if_bn=False, dnn_if_ln=False,
                 dnn_l2=0.0, dnn_initializer=None,
                 seed=None):
        super().__init__()
        if im_mode not in ('IM', 'DeepIM'):
            raise MLGBError
        if not (1 <= im_order <= 5):
            raise MLGBError

        self.im_mode = im_mode
        self.im_order = im_order
        self.im_l2 = im_l2
        self.im_initializer = InitializerLayer(
            initializer=im_initializer,
            activation=None,
            seed=seed,
        ).get()
        if self.im_mode == 'DeepIM':
            self.dnn_fn = DeepNeuralNetworkLayer(
                dnn_hidden_units=dnn_hidden_units,
                dnn_activation=dnn_activation,
                dnn_dropout=dnn_dropout,
                dnn_if_bn=dnn_if_bn,
                dnn_if_ln=dnn_if_ln,
                dnn_l2=dnn_l2,
                dnn_initializer=dnn_initializer,
                seed=seed,
            )
        self.flatten_fn = Flatten()

    def build(self, input_shape):
        if input_shape.rank != 3:
            raise MLGBError

        self.im_weight = self.add_weight(
            name='im_weight',
            shape=input_shape.as_list()[1:],
            initializer=self.im_initializer,
            regularizer=L1L2(0.0, self.im_l2),
            trainable=True,
        )
        self.built = True
        return

    @tf.function
    def call(self, inputs):
        x_3d = inputs
        x_3d = x_3d * self.im_weight

        p = self.get_x_list(x_3d, n=self.im_order)
        x_pool = [self.order_fn(p, i) for i in range(self.im_order)]
        x_im = tf.concat(x_pool, axis=1)  # (b, order * e)

        if self.im_mode == 'DeepIM':
            x_2d = self.flatten_fn(x_3d)
            x_dnn = self.dnn_fn(x_2d)
            x_im = tf.concat([x_im, x_dnn], axis=1)
        return x_im

    def order_fn(self, p, i):
        order_fn_list = [
            lambda p: p[0],
            lambda p: (p[0] ** 2 - p[1]) / 2,
            lambda p: (p[0] ** 3 - 3 * p[0] * p[1] + 2 * p[2]) / 6,
            lambda p: (p[0] ** 4 - 6 * p[0] ** 2 * p[1] + 3 * p[1] ** 2 + 8 * p[0] * p[2] - 6 * p[3]) / 24,
            lambda p: (p[0] ** 5 - 10 * p[0] ** 3 * p[1] + 20 * p[0] ** 2 * p[2] - 30 * p[0] * p[3]
                       - 20 * p[1] * p[2] + 15 * p[0] * p[1] ** 2 + 24 * p[4]) / 120,
        ]
        return order_fn_list[i](p)

    def get_x_list(self, x, n):
        x = tf.stack([x] * n, axis=1)
        x = tf.math.cumprod(x, axis=1)
        x = tf.reduce_sum(x, axis=2, keepdims=False)
        x_list = tf.unstack(x, n, axis=1)  # (b, e)
        return x_list


class AllInputFactorizationMachineLayer(tf.keras.layers.Layer):
    def __init__(self, ifm_mode_if_dual=False,
                 fen_hidden_units=(64, 32), fen_activation='relu', fen_dropout=0.0,
                 fen_if_bn=False, fen_if_ln=False, fen_l2=0.0, fen_initializer=None,
                 trm_mha_head_num=8, trm_mha_head_dim=32,
                 trm_mha_if_mask=False, trm_mha_l2=0.0, trm_mha_initializer=None, trm_residual_dropout=0.0,
                 fbi_mode='FM3D', fbi_unit=32, fbi_l2=0.0, fbi_initializer=None,
                 fbi_hofm_order=3, fbi_afm_activation='relu', fbi_afm_dropout=0.0,
                 linear_l1=0.0, linear_l2=0.0, linear_initializer=None,
                 seed=None):
        super().__init__()
        if fbi_mode == 'FM':
            raise MLGBError

        self.ifm_mode_if_dual = ifm_mode_if_dual

        if self.ifm_mode_if_dual:
            self.trm_fn = InteractingLayer(
                trm_mha_head_num=trm_mha_head_num,
                trm_mha_head_dim=trm_mha_head_dim,
                trm_mha_if_mask=trm_mha_if_mask,
                trm_mha_l2=trm_mha_l2,
                trm_mha_initializer=trm_mha_initializer,
                trm_residual_dropout=trm_residual_dropout,
                seed=seed,
            )
        self.fen_fn = FactorEstimatingNetworkLayer(
            ifm_mode_if_dual=ifm_mode_if_dual,
            dnn_hidden_units=fen_hidden_units,
            dnn_activation=fen_activation,
            dnn_dropout=fen_dropout,
            dnn_if_bn=fen_if_bn,
            dnn_if_ln=fen_if_ln,
            dnn_l2=fen_l2,
            dnn_initializer=fen_initializer,
            seed=seed,
        )
        self.fm_fn = AllFactorizationMachineLayer(
            fbi_mode=fbi_mode,
            fbi_unit=fbi_unit,
            fbi_l2=fbi_l2,
            fbi_initializer=fbi_initializer,
            fbi_hofm_order=fbi_hofm_order,
            fbi_afm_activation=fbi_afm_activation,
            fbi_afm_dropout=fbi_afm_dropout,
            linear_l1=linear_l1,
            linear_l2=linear_l2,
            linear_initializer=linear_initializer,
            seed=seed,
        )

    def build(self, input_shape):
        if len(input_shape) != 2:
            raise MLGBError
        if input_shape[0].rank != 2:
            raise MLGBError
        if input_shape[1].rank != 3:
            raise MLGBError

        self.built = True
        return

    @tf.function
    def call(self, inputs):
        dense_2d_tensor, embed_3d_tensor = inputs

        w_bit = self.fen_fn(embed_3d_tensor)  # (b, f, 1)
        if self.ifm_mode_if_dual:
            w_vec = self.trm_fn([embed_3d_tensor, embed_3d_tensor])  # (b, f, e)
            w_ifm = w_vec + w_bit
        else:
            w_ifm = w_bit

        embed_3d_tensor = embed_3d_tensor * w_ifm
        ifm_outputs = self.fm_fn([dense_2d_tensor, embed_3d_tensor])
        return ifm_outputs


class InputFactorizationMachineLayer(AllInputFactorizationMachineLayer):
    def __init__(self, ifm_mode_if_dual=False,
                 fen_hidden_units=(64, 32), fen_activation='relu', fen_dropout=0.0,
                 fen_if_bn=False, fen_if_ln=False, fen_l2=0.0, fen_initializer=None,
                 trm_mha_head_num=8, trm_mha_head_dim=32,
                 trm_mha_if_mask=False, trm_mha_l2=0.0, trm_mha_initializer=None, trm_residual_dropout=0.0,
                 fbi_mode='FM3D', fbi_unit=32, fbi_l2=0.0, fbi_initializer=None,
                 fbi_hofm_order=3, fbi_afm_activation='relu', fbi_afm_dropout=0.0,
                 linear_l1=0.0, linear_l2=0.0, linear_initializer=None,
                 seed=None):
        super().__init__(
            ifm_mode_if_dual,
            fen_hidden_units, fen_activation, fen_dropout,
            fen_if_bn, fen_if_ln, fen_l2, fen_initializer,
            trm_mha_head_num, trm_mha_head_dim,
            trm_mha_if_mask, trm_mha_l2, trm_mha_initializer, trm_residual_dropout,
            fbi_mode, fbi_unit, fbi_l2, fbi_initializer,
            fbi_hofm_order, fbi_afm_activation, fbi_afm_dropout,
            linear_l1, linear_l2, linear_initializer,
            seed,
        )
        if ifm_mode_if_dual:
            raise MLGBError


class DualInputFactorizationMachineLayer(AllInputFactorizationMachineLayer):
    def __init__(self, ifm_mode_if_dual=True,
                 fen_hidden_units=(64, 32), fen_activation='relu', fen_dropout=0.0,
                 fen_if_bn=False, fen_if_ln=False, fen_l2=0.0, fen_initializer=None,
                 trm_mha_head_num=8, trm_mha_head_dim=32,
                 trm_mha_if_mask=False, trm_mha_l2=0.0, trm_mha_initializer=None, trm_residual_dropout=0.0,
                 fbi_mode='FM3D', fbi_unit=32, fbi_l2=0.0, fbi_initializer=None,
                 fbi_hofm_order=3, fbi_afm_activation='relu', fbi_afm_dropout=0.0,
                 linear_l1=0.0, linear_l2=0.0, linear_initializer=None,
                 seed=None):
        super().__init__(
            ifm_mode_if_dual,
            fen_hidden_units, fen_activation, fen_dropout,
            fen_if_bn, fen_if_ln, fen_l2, fen_initializer,
            trm_mha_head_num, trm_mha_head_dim,
            trm_mha_if_mask, trm_mha_l2, trm_mha_initializer, trm_residual_dropout,
            fbi_mode, fbi_unit, fbi_l2, fbi_initializer,
            fbi_hofm_order, fbi_afm_activation, fbi_afm_dropout,
            linear_l1, linear_l2, linear_initializer,
            seed,
        )
        if not ifm_mode_if_dual:
            raise MLGBError


class FactorizationMachineNeuralNetworkLayer(tf.keras.layers.Layer):
    def __init__(self, fbi_mode='FM', fbi_unit=16, fbi_l2=0.0, fbi_initializer=None,
                 fbi_hofm_order=3, fbi_afm_activation='relu', fbi_afm_dropout=0.0,
                 linear_l1=0.0, linear_l2=0.0, linear_initializer=None,
                 dnn_hidden_units=(64, 32), dnn_activation='tanh', dnn_dropout=0.0, dnn_if_bn=False, dnn_if_ln=False,
                 dnn_l2=0.0, dnn_initializer=None,
                 seed=None):
        super().__init__()
        self.fm_fn = AllFactorizationMachineLayer(
            fbi_mode=fbi_mode,
            fbi_unit=fbi_unit,
            fbi_l2=fbi_l2,
            fbi_initializer=fbi_initializer,
            fbi_hofm_order=fbi_hofm_order,
            fbi_afm_activation=fbi_afm_activation,
            fbi_afm_dropout=fbi_afm_dropout,
            linear_l1=linear_l1,
            linear_l2=linear_l2,
            linear_initializer=linear_initializer,
            seed=seed,
        )
        self.dnn_fn = DeepNeuralNetworkLayer(
            dnn_hidden_units=dnn_hidden_units,
            dnn_activation=dnn_activation,
            dnn_dropout=dnn_dropout,
            dnn_if_bn=dnn_if_bn,
            dnn_if_ln=dnn_if_ln,
            dnn_l2=dnn_l2,
            dnn_initializer=dnn_initializer,
            seed=seed,
        )

    def build(self, input_shape):
        if len(input_shape) != 2:
            raise MLGBError
        if input_shape[0].rank != 2:
            raise MLGBError
        if input_shape[1].rank not in (2, 3):
            raise MLGBError

        self.built = True
        return

    @tf.function
    def call(self, inputs):
        linear_tensor, embed_3d_tensor = inputs

        fm_outputs = self.fm_fn([linear_tensor, embed_3d_tensor])
        fnn_outputs = self.dnn_fn(fm_outputs)
        return fnn_outputs


class ProductNeuralNetworkLayer(tf.keras.layers.Layer):
    def __init__(self, fbi_mode='PNN:both', fbi_l2=0.0, fbi_initializer=None,
                 fbi_hofm_order=3, fbi_afm_activation='relu', fbi_afm_dropout=0.0,
                 linear_l1=0.0, linear_l2=0.0, linear_initializer=None,
                 dnn_hidden_units=(64, 32), dnn_activation='relu', dnn_dropout=0.0, dnn_if_bn=False, dnn_if_ln=False,
                 dnn_l2=0.0, dnn_initializer=None,
                 seed=None):
        super().__init__()
        self.dnn_length = len(dnn_hidden_units)
        self.dnn_hidden_units = dnn_hidden_units
        self.dnn_dropouts = [dnn_dropout] * self.dnn_length if isinstance(dnn_dropout, float) else dnn_dropout
        self.dnn_bns = [dnn_if_bn] * self.dnn_length if isinstance(dnn_if_bn, bool) else dnn_if_bn
        self.dnn_lns = [dnn_if_ln] * self.dnn_length if isinstance(dnn_if_ln, bool) else dnn_if_ln

        self.dnn_first_layer = [
            BatchNormalization(axis=1) if self.dnn_bns[0] else IdentityLayer(),
            LayerNormalization(axis=1) if self.dnn_lns[0] else IdentityLayer(),
            Dropout(self.dnn_dropouts[0]),
            ActivationLayer(dnn_activation),
        ]
        self.fm_fn = AllFactorizationMachineLayer(
            fbi_mode=fbi_mode,
            fbi_unit=self.dnn_hidden_units[0],
            fbi_l2=fbi_l2,
            fbi_initializer=fbi_initializer,
            fbi_hofm_order=fbi_hofm_order,
            fbi_afm_activation=fbi_afm_activation,
            fbi_afm_dropout=fbi_afm_dropout,
            linear_l1=linear_l1,
            linear_l2=linear_l2,
            linear_initializer=linear_initializer,
            seed=seed,
        )
        self.dnn_fn = DeepNeuralNetworkLayer(
            dnn_hidden_units=self.dnn_hidden_units[1:],
            dnn_dropout=self.dnn_dropouts[1:],
            dnn_if_bn=self.dnn_bns[1:],
            dnn_if_ln=self.dnn_lns[1:],
            dnn_activation=dnn_activation,
            dnn_l2=dnn_l2,
            dnn_initializer=dnn_initializer,
            seed=seed,
        )

    def build(self, input_shape):
        if len(input_shape) != 2:
            raise MLGBError
        if input_shape[0].rank != 2:
            raise MLGBError
        if input_shape[1].rank != 3:
            raise MLGBError

        self.built = True
        return

    @tf.function
    def call(self, inputs):
        dense_2d_tensor, embed_3d_tensor = inputs

        x = self.fm_fn([dense_2d_tensor, embed_3d_tensor])
        for fn in self.dnn_first_layer:
            x = fn(x)
        x = self.dnn_fn(x)
        return x


class ProductNetworkInNetworkLayer(tf.keras.layers.Layer):
    def __init__(self, pin_parallel_num=4,
                 fbi_mode='PNN:both', fbi_unit=16, fbi_l2=0.0, fbi_initializer=None,
                 fbi_hofm_order=3, fbi_afm_activation='relu', fbi_afm_dropout=0.0,
                 linear_l1=0.0, linear_l2=0.0, linear_initializer=None,
                 dnn_hidden_units=(64, 32), dnn_activation='relu', dnn_dropout=0.0, dnn_if_bn=False, dnn_if_ln=True,
                 dnn_l2=0.0, dnn_initializer=None,
                 seed=None):
        super().__init__()
        if fbi_mode not in FBIModeList:
            raise MLGBError

        self.pin_parallel_num = pin_parallel_num

        self.fm_fn = AllFactorizationMachineLayer(
            fbi_mode=fbi_mode,
            fbi_unit=fbi_unit,
            fbi_l2=fbi_l2,
            fbi_initializer=fbi_initializer,
            fbi_hofm_order=fbi_hofm_order,
            fbi_afm_activation=fbi_afm_activation,
            fbi_afm_dropout=fbi_afm_dropout,
            linear_l1=linear_l1,
            linear_l2=linear_l2,
            linear_initializer=linear_initializer,
            seed=seed,
        )
        self.dnn_parallel_fn = DNN3dParallelLayer(
            dnn_if_output2d=True,
            dnn_hidden_units=dnn_hidden_units,
            dnn_activation=dnn_activation,
            dnn_dropout=dnn_dropout,
            dnn_if_bn=dnn_if_bn,
            dnn_if_ln=dnn_if_ln,  #
            dnn_l2=dnn_l2,
            dnn_initializer=dnn_initializer,
            seed=seed,
        )

    def build(self, input_shape):
        if len(input_shape) != 2:
            raise MLGBError
        if input_shape[0].rank != 2:
            raise MLGBError
        if input_shape[1].rank != 3:
            raise MLGBError

        self.built = True
        return

    @tf.function
    def call(self, inputs):
        dense_2d_tensor, embed_3d_tensor = inputs

        x = self.fm_fn([dense_2d_tensor, embed_3d_tensor])
        x = tf.stack([x] * self.pin_parallel_num, axis=1)
        x = self.dnn_parallel_fn(x)
        return x


class OperationNeuralNetworkLayer(tf.keras.layers.Layer):
    def __init__(self, fbi_mode='FFM', fbi_unit=16, fbi_l2=0.0, fbi_initializer=None,
                 fbi_hofm_order=3, fbi_afm_activation='relu', fbi_afm_dropout=0.0,
                 dnn_hidden_units=(64, 32), dnn_activation='relu', dnn_dropout=0.0, dnn_if_bn=True, dnn_if_ln=False,
                 dnn_l2=0.0, dnn_initializer=None,
                 seed=None):
        super().__init__()
        self.fbi_fn = AllFieldBinaryInteractionLayer(
            fbi_mode=fbi_mode,
            fbi_unit=fbi_unit,
            fbi_l2=fbi_l2,
            fbi_initializer=fbi_initializer,
            fbi_hofm_order=fbi_hofm_order,
            fbi_afm_activation=fbi_afm_activation,
            fbi_afm_dropout=fbi_afm_dropout,
            seed=seed,
        )
        self.dnn_fn = DeepNeuralNetworkLayer(
            dnn_hidden_units=dnn_hidden_units,
            dnn_dropout=dnn_dropout,
            dnn_if_bn=dnn_if_bn,
            dnn_if_ln=dnn_if_ln,
            dnn_activation=dnn_activation,
            dnn_l2=dnn_l2,
            dnn_initializer=dnn_initializer,
            seed=seed,
        )
        self.flatten_fn = Flatten()

    def build(self, input_shape):
        if input_shape.rank != 3:
            raise MLGBError

        self.built = True
        return

    @tf.function
    def call(self, inputs):
        embed_3d_tensor = inputs
        embed_2d_tensor = self.flatten_fn(embed_3d_tensor)

        fbi_outputs = self.fbi_fn(embed_3d_tensor)
        ffm_outputs = tf.concat([fbi_outputs, embed_2d_tensor], axis=1)
        nffm_outputs = self.dnn_fn(ffm_outputs)
        return nffm_outputs


class AdaptiveFactorizationNetworkLayer(tf.keras.layers.Layer):
    def __init__(self, afn_mode_if_ensemble=True,
                 ltl_clip_min=1e-4, ltl_unit=32, ltl_l2=0.0, ltl_initializer=None,
                 dnn_hidden_units=(64, 32), dnn_activation='relu', dnn_dropout=0.0,
                 dnn_if_bn=True, dnn_if_ln=False, dnn_l2=0.0, dnn_initializer=None,
                 ensemble_dnn_hidden_units=(64, 32), ensemble_dnn_activation='relu', ensemble_dnn_dropout=0.0,
                 ensemble_dnn_if_bn=True, ensemble_dnn_if_ln=False, ensemble_dnn_l2=0.0, ensemble_dnn_initializer=None,
                 seed=None):
        super().__init__()
        self.afn_mode_if_ensemble = afn_mode_if_ensemble

        self.ltl_fn = LogarithmicTransformationLayer(
            ltl_clip_min=ltl_clip_min,
            ltl_unit=ltl_unit,
            ltl_l2=ltl_l2,
            ltl_initializer=ltl_initializer,
            seed=seed,
        )
        self.dnn_fn = DeepNeuralNetworkLayer(
            dnn_hidden_units=dnn_hidden_units,
            dnn_dropout=dnn_dropout,
            dnn_if_bn=dnn_if_bn,
            dnn_if_ln=dnn_if_ln,
            dnn_activation=dnn_activation,
            dnn_l2=dnn_l2,
            dnn_initializer=dnn_initializer,
            seed=seed,
        )
        if self.afn_mode_if_ensemble:
            self.ensemble_dnn_fn = DeepNeuralNetworkLayer(
                dnn_hidden_units=ensemble_dnn_hidden_units,
                dnn_dropout=ensemble_dnn_dropout,
                dnn_if_bn=ensemble_dnn_if_bn,
                dnn_if_ln=ensemble_dnn_if_ln,
                dnn_activation=ensemble_dnn_activation,
                dnn_l2=ensemble_dnn_l2,
                dnn_initializer=ensemble_dnn_initializer,
                seed=seed,
            )
        self.flatten_fn = Flatten()

    def build(self, input_shape):
        if input_shape.rank != 3:
            raise MLGBError

        self.built = True
        return

    @tf.function
    def call(self, inputs):
        embed_3d_tensor = inputs
        embed_2d_tensor = self.flatten_fn(embed_3d_tensor)

        ltl_outputs = self.ltl_fn(embed_3d_tensor)
        afn_outputs = self.dnn_fn(ltl_outputs)

        if self.afn_mode_if_ensemble:
            dnn_outputs = self.ensemble_dnn_fn(embed_2d_tensor)
            afn_outputs = tf.concat([afn_outputs, dnn_outputs], axis=1)
        return afn_outputs


class NeuralFactorizationMachineLayer(tf.keras.layers.Layer):
    def __init__(self, fbi_mode='FM3D', fbi_unit=16, fbi_l2=0.0, fbi_initializer=None,
                 fbi_hofm_order=3, fbi_afm_activation='relu', fbi_afm_dropout=0.0,
                 linear_l1=0.0, linear_l2=0.0, linear_initializer=None,
                 dnn_hidden_units=(64, 32), dnn_activation='relu', dnn_dropout=0.0, dnn_if_bn=False, dnn_if_ln=False,
                 dnn_l2=0.0, dnn_initializer=None,
                 seed=None):
        super().__init__()
        self.linear_fn = LinearLayer(
            linear_l1=linear_l1,
            linear_l2=linear_l2,
            linear_initializer=linear_initializer,
            seed=seed,
        )
        self.fbi_fn = AllFieldBinaryInteractionLayer(
            fbi_mode=fbi_mode,
            fbi_unit=fbi_unit,
            fbi_l2=fbi_l2,
            fbi_initializer=fbi_initializer,
            fbi_hofm_order=fbi_hofm_order,
            fbi_afm_activation=fbi_afm_activation,
            fbi_afm_dropout=fbi_afm_dropout,
            seed=seed,
        )
        self.dnn_fn = DeepNeuralNetworkLayer(
            dnn_hidden_units=dnn_hidden_units,
            dnn_activation=dnn_activation,
            dnn_dropout=dnn_dropout,
            dnn_if_bn=dnn_if_bn,
            dnn_if_ln=dnn_if_ln,
            dnn_l2=dnn_l2,
            dnn_initializer=dnn_initializer,
            seed=seed,
        )

    def build(self, input_shape):
        if len(input_shape) != 2:
            raise MLGBError
        if input_shape[0].rank != 2:
            raise MLGBError
        if input_shape[1].rank != 3:
            raise MLGBError

        self.built = True
        return

    @tf.function
    def call(self, inputs):
        dense_2d_tensor, embed_3d_tensor = inputs

        linear_outputs = self.linear_fn(dense_2d_tensor)
        fbi_tensor = self.fbi_fn(embed_3d_tensor)
        dnn_outputs = self.dnn_fn(fbi_tensor)
        nfm_outputs = tf.concat([linear_outputs, dnn_outputs], axis=1)
        return nfm_outputs


class WideDeepLearningLayer(tf.keras.layers.Layer):
    def __init__(self, fbi_mode='FM', fbi_unit=32, fbi_l2=0.0, fbi_initializer=None,
                 fbi_hofm_order=3, fbi_afm_activation='relu', fbi_afm_dropout=0.0,
                 dnn_hidden_units=(64, 32), dnn_activation='relu', dnn_dropout=0.0, dnn_if_bn=False, dnn_if_ln=False,
                 dnn_l2=0.0, dnn_initializer=None,
                 seed=None):
        super().__init__()
        self.fbi_fn = AllFieldBinaryInteractionLayer(
            fbi_mode=fbi_mode,
            fbi_unit=fbi_unit,
            fbi_l2=fbi_l2,
            fbi_initializer=fbi_initializer,
            fbi_hofm_order=fbi_hofm_order,
            fbi_afm_activation=fbi_afm_activation,
            fbi_afm_dropout=fbi_afm_dropout,
            seed=seed,
        )
        self.dnn_fn = DeepNeuralNetworkLayer(
            dnn_hidden_units=dnn_hidden_units,
            dnn_activation=dnn_activation,
            dnn_dropout=dnn_dropout,
            dnn_if_bn=dnn_if_bn,
            dnn_if_ln=dnn_if_ln,
            dnn_l2=dnn_l2,
            dnn_initializer=dnn_initializer,
            seed=seed,
        )

    def build(self, input_shape):
        if len(input_shape) != 2:
            raise MLGBError
        if input_shape[0].rank != 2:
            raise MLGBError
        if input_shape[1].rank not in (2, 3):
            raise MLGBError

        self.built = True
        return

    @tf.function
    def call(self, inputs):
        dense_2d_tensor, embed_3d_tensor = inputs

        wide_outputs = self.fbi_fn(embed_3d_tensor)
        deep_outputs = self.dnn_fn(dense_2d_tensor)
        wdl_outputs = tf.concat([wide_outputs, deep_outputs], axis=1)
        return wdl_outputs


class DeepFactorizationMachineLayer(tf.keras.layers.Layer):
    def __init__(self, fbi_mode='FM', fbi_unit=32, fbi_l2=0.0, fbi_initializer=None,
                 fbi_hofm_order=3, fbi_afm_activation='relu', fbi_afm_dropout=0.0,
                 linear_l1=0.0, linear_l2=0.0, linear_initializer=None,
                 dnn_hidden_units=(64, 32), dnn_activation='relu', dnn_dropout=0.0, dnn_if_bn=False, dnn_if_ln=False,
                 dnn_l2=0.0, dnn_initializer=None,
                 seed=None):
        super().__init__()
        self.fm_fn = AllFactorizationMachineLayer(
            fbi_mode=fbi_mode,
            fbi_unit=fbi_unit,
            fbi_l2=fbi_l2,
            fbi_initializer=fbi_initializer,
            fbi_hofm_order=fbi_hofm_order,
            fbi_afm_activation=fbi_afm_activation,
            fbi_afm_dropout=fbi_afm_dropout,
            linear_initializer=linear_initializer,
            linear_l1=linear_l1,
            linear_l2=linear_l2,
            seed=seed,
        )
        self.dnn_fn = DeepNeuralNetworkLayer(
            dnn_hidden_units=dnn_hidden_units,
            dnn_activation=dnn_activation,
            dnn_dropout=dnn_dropout,
            dnn_if_bn=dnn_if_bn,
            dnn_if_ln=dnn_if_ln,
            dnn_l2=dnn_l2,
            dnn_initializer=dnn_initializer,
            seed=seed,
        )

    def build(self, input_shape):
        if len(input_shape) != 2:
            raise MLGBError
        if input_shape[0].rank != 2:
            raise MLGBError
        if input_shape[1].rank not in (2, 3):
            raise MLGBError

        self.built = True
        return

    @tf.function
    def call(self, inputs):
        dense_2d_tensor, embed_3d_tensor = inputs

        fm_outputs = self.fm_fn([dense_2d_tensor, embed_3d_tensor])
        deep_outputs = self.dnn_fn(dense_2d_tensor)
        dfm_outputs = tf.concat([fm_outputs, deep_outputs], axis=1)
        return dfm_outputs


class DeepFieldEmbeddedFactorizationMachineLayer(tf.keras.layers.Layer):
    def __init__(self, fbi_mode='FEFM', fbi_unit=1, fbi_l2=0.0, fbi_initializer=None,
                 fbi_hofm_order=3, fbi_afm_activation='relu', fbi_afm_dropout=0.0,
                 linear_l1=0.0, linear_l2=0.0, linear_initializer=None,
                 dnn_hidden_units=(64, 32), dnn_activation='relu', dnn_dropout=0.0, dnn_if_bn=False, dnn_if_ln=False,
                 dnn_l2=0.0, dnn_initializer=None,
                 seed=None):
        super().__init__()
        self.fm_fn = AllFactorizationMachineLayer(
            fbi_mode=fbi_mode,
            fbi_unit=fbi_unit,
            fbi_l2=fbi_l2,
            fbi_initializer=fbi_initializer,
            fbi_hofm_order=fbi_hofm_order,
            fbi_afm_activation=fbi_afm_activation,
            fbi_afm_dropout=fbi_afm_dropout,
            linear_l1=linear_l1,
            linear_l2=linear_l2,
            linear_initializer=linear_initializer,
            seed=seed,
        )
        self.dnn_fn = DeepNeuralNetworkLayer(
            dnn_hidden_units=dnn_hidden_units,
            dnn_activation=dnn_activation,
            dnn_dropout=dnn_dropout,
            dnn_if_bn=dnn_if_bn,
            dnn_if_ln=dnn_if_ln,
            dnn_l2=dnn_l2,
            dnn_initializer=dnn_initializer,
            seed=seed,
        )

    def build(self, input_shape):
        if len(input_shape) != 2:
            raise MLGBError
        if input_shape[0].rank != 2:
            raise MLGBError
        if input_shape[1].rank not in (2, 3):
            raise MLGBError

        self.built = True
        return

    @tf.function
    def call(self, inputs):
        dense_2d_tensor, embed_3d_tensor = inputs

        fm_outputs = self.fm_fn([dense_2d_tensor, embed_3d_tensor])
        dnn_outputs = self.dnn_fn(fm_outputs)
        deepfefm_outputs = tf.concat([fm_outputs, dnn_outputs], axis=1)
        return deepfefm_outputs


class FieldLeveragedEmbeddingNetworkLayer(tf.keras.layers.Layer):
    def __init__(self, flen_group_indices=((0, 1), (2, 3), (4, 5, 6)),
                 fbi_fm_mode='FM3D', fbi_mf_mode='FwFM', fbi_unit=16, fbi_l2=0.0, fbi_initializer=None,
                 fbi_afm_activation='relu', fbi_afm_dropout=0.0,
                 linear_l1=0.0, linear_l2=0.0, linear_initializer=None,
                 dnn_hidden_units=(64, 32), dnn_activation='relu', dnn_dropout=0.0, dnn_if_bn=False, dnn_if_ln=False,
                 dnn_l2=0.0, dnn_initializer=None,
                 seed=None):
        super().__init__()
        self.linear_fn = LinearLayer(
            linear_initializer=linear_initializer,
            linear_l1=linear_l1,
            linear_l2=linear_l2,
            seed=seed,
        )
        self.dnn_fn = DeepNeuralNetworkLayer(
            dnn_hidden_units=dnn_hidden_units,
            dnn_activation=dnn_activation,
            dnn_dropout=dnn_dropout,
            dnn_if_bn=dnn_if_bn,
            dnn_if_ln=dnn_if_ln,
            dnn_l2=dnn_l2,
            dnn_initializer=dnn_initializer,
            seed=seed,
        )
        self.fwbi_fn = GroupedAllFieldWiseBinaryInteractionLayer(
            group_indices=flen_group_indices,
            fbi_fm_mode=fbi_fm_mode,
            fbi_mf_mode=fbi_mf_mode,
            fbi_unit=fbi_unit,
            fbi_l2=fbi_l2,
            fbi_initializer=fbi_initializer,
            fbi_afm_activation=fbi_afm_activation,
            fbi_afm_dropout=fbi_afm_dropout,
            seed=seed,
        )

    def build(self, input_shape):
        if len(input_shape) != 2:
            raise MLGBError
        if input_shape[0].rank != 2:
            raise MLGBError
        if input_shape[1].rank != 3:
            raise MLGBError

        self.built = True
        return

    @tf.function
    def call(self, inputs):
        dense_2d_tensor, embed_3d_tensor = inputs

        fwbi_outputs = self.fwbi_fn(embed_3d_tensor)
        dnn_ouputs = self.dnn_fn(dense_2d_tensor)
        linear_outputs = self.linear_fn(dense_2d_tensor)
        flen_outputs = tf.concat([fwbi_outputs, dnn_ouputs, linear_outputs], axis=1)
        return flen_outputs


class ConvolutionalClickPredictionModelLayer(tf.keras.layers.Layer):
    def __init__(self, cnn_filter_nums=(64, 32), cnn_kernel_sizes=(64, 32), cnn_activation='tanh',
                 cnn_l2=0.0, cnn_initializer=None,
                 dnn_hidden_units=(64, 32), dnn_activation='relu', dnn_dropout=0.0, dnn_if_bn=False, dnn_if_ln=False,
                 dnn_l2=0.0, dnn_initializer=None,
                 seed=None):
        super().__init__()
        self.p_min = 3
        if min(cnn_filter_nums) < self.p_min or min(cnn_kernel_sizes) < self.p_min:
            raise MLGBError

        self.cnn_length = len(cnn_filter_nums)

        self.cnn_fn_list = [
            ConvolutionalNeuralNetworkLayer(
                cnn_conv_mode='Conv2D',
                cnn_filter_nums=[cnn_filter_nums[i]],
                cnn_kernel_heights=[cnn_kernel_sizes[i]],
                cnn_kernel_widths=1,
                cnn_activation=cnn_activation,
                cnn_l2=cnn_l2,
                cnn_initializer=cnn_initializer,
                cnn_if_max_pool=False,
                seed=seed + i if isinstance(seed, int) else seed,
            )
            for i in range(self.cnn_length)
        ]
        self.dnn_fn = DeepNeuralNetworkLayer(
            dnn_hidden_units=dnn_hidden_units,
            dnn_activation=dnn_activation,
            dnn_dropout=dnn_dropout,
            dnn_if_bn=dnn_if_bn,
            dnn_if_ln=dnn_if_ln,
            dnn_l2=dnn_l2,
            dnn_initializer=dnn_initializer,
            seed=seed,
        )
        self.k_max_pooling_fn = KMaxPoolingLayer(pool_axis=1)
        self.flatten_fn = Flatten()

    def build(self, input_shape):
        if input_shape.rank != 3:
            raise MLGBError
        if input_shape[1] < self.p_min:  # fields_width
            raise MLGBError

        self.built = True
        return

    @tf.function
    def call(self, inputs):
        x = inputs

        x = tf.expand_dims(x, axis=3)  # (b, f, e, 1) == (b, h, w, i_c)
        for i in range(self.cnn_length):
            x = self.cnn_fn_list[i](x)  # (b, f, e, o_c)
            p = self.flexible_p(i=i+1, j=self.cnn_length, n=x.shape[1])  # n = flexible features width
            x = self.k_max_pooling_fn(x, k=p)

        x = self.flatten_fn(x)
        x = self.dnn_fn(x)
        return x

    def flexible_p(self, i, j, n):
        if i < 1 or j < 1 or i > j:
            raise MLGBError

        p = max(int((1 - (i / j) ** (j - i)) * n), self.p_min)
        return p


class FeatureGenerationByConvolutionalNeuralNetworkLayer(tf.keras.layers.Layer):
    def __init__(self, cnn_filter_nums=(64, 32), cnn_kernel_sizes=(64, 32), cnn_pool_sizes=(2, 2),
                 cnn_activation='tanh', cnn_l2=0.0, cnn_initializer=None,
                 recomb_dnn_hidden_units=(64, 32), recomb_dnn_activation='tanh',
                 fbi_mode='PNN:inner_product', fbi_unit=32, fbi_l2=0.0, fbi_initializer=None,
                 fbi_hofm_order=3, fbi_afm_activation='relu', fbi_afm_dropout=0.0,
                 dnn_hidden_units=(64, 32), dnn_activation='relu', dnn_dropout=0.0, dnn_if_bn=False, dnn_if_ln=False,
                 dnn_l2=0.0, dnn_initializer=None,
                 seed=None):
        super().__init__()
        self.cnn_length = len(cnn_filter_nums)
        if not (self.cnn_length == len(cnn_kernel_sizes) == len(cnn_pool_sizes) == len(recomb_dnn_hidden_units)):
            raise MLGBError('self.cnn_length == len(cnn_kernel_sizes) == len(cnn_pool_sizes) == len(recomb_dnn_units)')
        if fbi_mode == 'FM':
            raise MLGBError

        self.cpr_fn_list = [
            [
                ConvolutionalNeuralNetworkLayer(
                    cnn_conv_mode='Conv2D',
                    cnn_filter_nums=[cnn_filter_nums[i]],
                    cnn_kernel_heights=[cnn_kernel_sizes[i]],
                    cnn_kernel_widths=1,
                    cnn_activation=cnn_activation,
                    cnn_l2=cnn_l2,
                    cnn_initializer=cnn_initializer,
                    cnn_if_max_pool=False,
                    seed=seed + i if isinstance(seed, int) else seed,
                ),  # (b, f, e, 1) -> (b, (f - p + 1) // p, e, cnn_filter_num)
                FlattenAxesLayer(axes=[1, 3]),  # (b, cnn_filter_num * (f - p + 1) // p, e)
                TransposeLayer(perm=[0, 2, 1]),
                DNN3dParallelLayer(
                    dnn_hidden_units=[recomb_dnn_hidden_units[i]],
                    dnn_activation=recomb_dnn_activation,
                    dnn_if_output2d=False,
                    seed=seed + i if isinstance(seed, int) else seed,
                ),
                TransposeLayer(perm=[0, 2, 1]),  # (b, recomb_dnn_hidden_unit, e)
            ]
            for i in range(self.cnn_length)
        ]
        self.fbi_fn = AllFieldBinaryInteractionLayer(
            fbi_mode=fbi_mode,
            fbi_unit=fbi_unit,
            fbi_l2=fbi_l2,
            fbi_initializer=fbi_initializer,
            fbi_hofm_order=fbi_hofm_order,
            fbi_afm_activation=fbi_afm_activation,
            fbi_afm_dropout=fbi_afm_dropout,
            seed=seed,
        )
        self.dnn_fn = DeepNeuralNetworkLayer(
            dnn_hidden_units=dnn_hidden_units,
            dnn_activation=dnn_activation,
            dnn_dropout=dnn_dropout,
            dnn_if_bn=dnn_if_bn,
            dnn_if_ln=dnn_if_ln,
            dnn_l2=dnn_l2,
            dnn_initializer=dnn_initializer,
            seed=seed,
        )
        self.flatten_fn = Flatten()

    def build(self, input_shape):
        if len(input_shape) != 2:
            raise MLGBError
        if input_shape[0].rank != 2:
            raise MLGBError
        if input_shape[1].rank != 3:
            raise MLGBError

        self.built = True
        return

    @tf.function
    def call(self, inputs):
        dense_2d_tensor, embed_3d_tensor = inputs

        cnn_3d_outputs = self.cpr_fn(embed_3d_tensor)
        cnn_2d_outputs = self.flatten_fn(cnn_3d_outputs)

        all_3d_features = tf.concat([embed_3d_tensor, cnn_3d_outputs], axis=1)
        fbi_outputs = self.fbi_fn(all_3d_features)

        all_dense_features = tf.concat([dense_2d_tensor, cnn_2d_outputs, fbi_outputs], axis=1)
        fgcnn_outputs = self.dnn_fn(all_dense_features)
        return fgcnn_outputs

    def cpr_fn(self, x):  # Conv_Pool_Recombination
        x_list = []
        for cpr_fn in self.cpr_fn_list:
            x_i = tf.expand_dims(x, axis=3)
            for _fn in cpr_fn:
                x_i = _fn(x_i)
            x_list.append(x_i)
        x = tf.concat(x_list, axis=1)
        return x


class ExtremeDeepFactorizationMachineLayer(tf.keras.layers.Layer):
    def __init__(self, cin_interaction_num=4, cin_interaction_ratio=1.0,
                 cnn_filter_num=64, cnn_kernel_size=64, cnn_activation='relu',
                 cnn_l2=0.0, cnn_initializer=None,
                 linear_l1=0.0, linear_l2=0.0, linear_initializer=None,
                 dnn_hidden_units=(64, 32), dnn_activation='relu', dnn_dropout=0.0, dnn_if_bn=False, dnn_if_ln=False,
                 dnn_l2=0.0, dnn_initializer=None,
                 seed=None):
        super().__init__()
        if not (0.5 <= cin_interaction_ratio <= 1.0):
            raise MLGBError('0.5 <= cin_interaction_ratio <= 1.0')

        self.cin_fn = CompressedInteractionNetworkLayer(
            cin_interaction_num=cin_interaction_num,
            cin_interaction_ratio=cin_interaction_ratio,
            cnn_filter_num=cnn_filter_num,
            cnn_kernel_size=cnn_kernel_size,
            cnn_activation=cnn_activation,
            cnn_l2=cnn_l2,
            cnn_initializer=cnn_initializer,
            seed=seed,
        )
        self.linear_fn = LinearLayer(
            linear_initializer=linear_initializer,
            linear_l1=linear_l1,
            linear_l2=linear_l2,
            seed=seed,
        )
        self.dnn_fn = DeepNeuralNetworkLayer(
            dnn_hidden_units=dnn_hidden_units,
            dnn_activation=dnn_activation,
            dnn_dropout=dnn_dropout,
            dnn_if_bn=dnn_if_bn,
            dnn_if_ln=dnn_if_ln,
            dnn_l2=dnn_l2,
            dnn_initializer=dnn_initializer,
            seed=seed,
        )

    def build(self, input_shape):
        if len(input_shape) != 2:
            raise MLGBError
        if input_shape[0].rank != 2:
            raise MLGBError
        if input_shape[1].rank != 3:
            raise MLGBError

        self.built = True
        return

    @tf.function
    def call(self, inputs):
        dense_2d_tensor, embed_3d_tensor = inputs

        linear_outputs = self.linear_fn(dense_2d_tensor)
        dnn_outputs = self.dnn_fn(dense_2d_tensor)
        cin_outputs = self.cin_fn(embed_3d_tensor)
        xdeepfm_outputs = tf.concat([linear_outputs, cin_outputs, dnn_outputs], axis=1)
        return xdeepfm_outputs


class FeatureImportanceBilinearInteractionNetworkLayer(tf.keras.layers.Layer):
    def __init__(self, sen_pool_mode='Pooling:average', sen_reduction_factor=2, sen_activation='relu',
                 sen_l2=0.0, sen_initializer=None,
                 fbi_mode='Bilinear:field_interaction', fbi_unit=16, fbi_l2=0.0, fbi_initializer=None,
                 fbi_hofm_order=3, fbi_afm_activation='relu', fbi_afm_dropout=0.0,
                 dnn_hidden_units=(64, 32), dnn_activation='relu', dnn_dropout=0.0, dnn_if_bn=False, dnn_if_ln=False,
                 dnn_l2=0.0, dnn_initializer=None,
                 seed=None):
        super().__init__()
        self.sen_fn = SqueezeExcitationNetworkLayer(
            sen_pool_mode=sen_pool_mode,
            sen_reduction_factor=sen_reduction_factor,
            sen_activation=sen_activation,
            sen_l2=sen_l2,
            sen_initializer=sen_initializer,
            seed=seed,
        )
        self.fbi_fn_list = [
            AllFieldBinaryInteractionLayer(
                fbi_mode=fbi_mode,
                fbi_unit=fbi_unit,
                fbi_l2=fbi_l2,
                fbi_initializer=fbi_initializer,
                fbi_hofm_order=fbi_hofm_order,
                fbi_afm_activation=fbi_afm_activation,
                fbi_afm_dropout=fbi_afm_dropout,
                seed=seed + i if isinstance(seed, int) else seed,
            )
            for i in range(2)
        ]
        self.dnn_fn = DeepNeuralNetworkLayer(
            dnn_hidden_units=dnn_hidden_units,
            dnn_activation=dnn_activation,
            dnn_dropout=dnn_dropout,
            dnn_if_bn=dnn_if_bn,
            dnn_if_ln=dnn_if_ln,
            dnn_l2=dnn_l2,
            dnn_initializer=dnn_initializer,
            seed=seed,
        )

    def build(self, input_shape):
        if len(input_shape) != 2:
            raise MLGBError
        if input_shape[0].rank != 2:
            raise MLGBError
        if input_shape[1].rank != 3:
            raise MLGBError

        self.built = True
        return

    @tf.function
    def call(self, inputs):
        dense_tensor, embed_3d_tensor = inputs

        fbi_tensor = self.fbi_fn_list[0](embed_3d_tensor)
        sen_tensor = self.sen_fn(embed_3d_tensor)
        fbi_sen_tensor = self.fbi_fn_list[1](sen_tensor)
        dnn_tensor = tf.concat([dense_tensor, fbi_tensor, fbi_sen_tensor], axis=1)

        fbn_outputs = self.dnn_fn(dnn_tensor)
        return fbn_outputs


class AutomaticFeatureInteractionLearningLayer(tf.keras.layers.Layer):
    def __init__(self, trm_layer_num=1, trm_mha_head_num=8, trm_mha_head_dim=32,
                 trm_mha_if_mask=True, trm_mha_l2=0.0, trm_mha_initializer=None,
                 trm_residual_dropout=0.0, seed=None):
        super().__init__()
        self.trm_layer_num = trm_layer_num

        self.trm_fn_list = [
            InteractingLayer(
                trm_mha_head_num=trm_mha_head_num,
                trm_mha_head_dim=trm_mha_head_dim,
                trm_mha_if_mask=trm_mha_if_mask,
                trm_mha_l2=trm_mha_l2,
                trm_mha_initializer=trm_mha_initializer,
                trm_residual_dropout=trm_residual_dropout,
                seed=seed + i if isinstance(seed, int) else seed,
            ) for i in range(trm_layer_num)
        ]
        self.flatten_fn = Flatten()

    def build(self, input_shape):
        if len(input_shape) != 2:
            raise MLGBError
        if not (input_shape[0].rank == input_shape[1].rank == 3):
            raise MLGBError
        if input_shape[0] != input_shape[1]:
            raise MLGBError

        self.built = True
        return

    @tf.function
    def call(self, inputs):
        x_q, x_k = inputs

        x = self.trm_fn_list[0]([x_q, x_k])
        if self.trm_layer_num > 1:
            for trm_fn in self.trm_fn_list[1:]:
                x = trm_fn([x, x])

        x = self.flatten_fn(x)
        return x









# coding=utf-8
# author=uliontse

"""
Copyright 2020 UlionTse

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

   http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""


from mlgb.torch.configs import torch
from mlgb.torch.functions import SeedLayer
from mlgb.torch.agg import (
    torch_ranking_models_map,
    torch_matching_models_map,
    torch_mtl_models_map,
)
from mlgb.tf.agg import (
    tf_ranking_models_map,
    tf_matching_models_map,
    tf_mtl_models_map,
)
from mlgb.error import MLGBError, MLGBWarning


class Tse:
    def __init__(self):
        self.tf_ranking_models_map = tf_ranking_models_map
        self.tf_matching_models_map = tf_matching_models_map
        self.tf_mtl_models_map = tf_mtl_models_map
        self.torch_ranking_models_map = torch_ranking_models_map
        self.torch_matching_models_map = torch_matching_models_map
        self.torch_mtl_models_map = torch_mtl_models_map
        self.all_models_map = {
            'tf_ranking_models_map': self.tf_ranking_models_map,
            'tf_matching_models_map': self.tf_matching_models_map,
            'tf_mtl_models_map': self.tf_mtl_models_map,
            'torch_ranking_models_map': self.torch_ranking_models_map,
            'torch_matching_models_map': self.torch_matching_models_map,
            'torch_mtl_models_map': self.torch_mtl_models_map,
        }
        self.ranking_models = list(self.tf_ranking_models_map.keys())
        self.matching_models = list(self.tf_matching_models_map.keys())
        self.mtl_models = list(self.tf_mtl_models_map.keys())
        self.all_models = list(set(self.ranking_models + self.matching_models + self.mtl_models))

    def get_short_lang(self, lang='TensorFlow'):
        if lang not in ('TensorFlow', 'PyTorch', 'tf', 'torch'):
            raise MLGBError('lang')
        return 'tf' if lang in ('TensorFlow', 'tf') else 'torch'

    def find_model(self, model_name='LR', aim='ranking', lang='TensorFlow'):
        if aim not in ('ranking', 'matching', 'mtl'):
            raise MLGBError('aim')
        if aim == 'ranking' and model_name not in self.ranking_models:
            raise MLGBError('model_name & aim')
        if aim == 'matching' and model_name not in self.matching_models:
            raise MLGBError('model_name & aim')
        if aim == 'mtl' and model_name not in self.mtl_models:
            raise MLGBError('model_name & aim')

        lang = self.get_short_lang(lang)
        model = self.all_models_map[f'{lang}_{aim}_models_map'][model_name]
        return model

    def get_model_help(self, model_name='LR', aim='ranking', lang='TensorFlow'):
        model = self.find_model(model_name=model_name, aim=aim, lang=lang)
        help(model)
        return

    def get_model(self, feature_names, model_name='LR', task='binary', aim='ranking', lang='TensorFlow',
                  device=None, seed=None, **kwargs):
        """
        :param feature_names: tuple(tuple(dict)), must. Embedding need vocabulary size and custom embed_dim of features.
        :param model_name: str, default 'LR'. Union[`mlgb.ranking_models`, `mlgb.matching_models`, `mlgb.mtl_models`]
        :param task: str, default 'binary'. Union['binary', 'regression', 'multiclass:{int}']
        :param aim: str, default 'ranking'. Union['ranking', 'matching', 'mtl']
        :param lang: str, default 'TensorFlow'. Union['TensorFlow', 'PyTorch', 'tf', 'torch']
        :param device: Optional[str, int], default None. Only for PyTorch.
        :param seed: Optional[int], default None.
        :param **kwargs: more model parameters by `mlgb.get_model_help(model_name)`.
        """

        model = self.find_model(model_name=model_name, aim=aim, lang=lang)
        if self.get_short_lang(lang) == 'tf':
            if device:
                MLGBWarning(f'tf_device:"{device}" needs to be set in global.')
            kwargs.update({'task': task, 'seed': seed})
        else:
            _ = SeedLayer(seed=seed).reset()
            if device is None:
                device = 'cuda' if torch.cuda.is_available() else 'cpu'
            kwargs.update({'task': task, 'device': device})
        return model(feature_names, **kwargs)


tse = Tse()
get_model = tse.get_model
get_model_help = tse.get_model_help
mtl_models = tse.mtl_models
ranking_models = tse.ranking_models
matching_models = tse.matching_models




# coding=utf-8
# author=UlionTse

import re
import pathlib
import setuptools


NAME = 'mlgb'
PACKAGE = 'mlgb'
AUTHOR = 'UlionTse'
AUTHOR_EMAIL = 'uliontse@outlook.com'
HOMEPAGE_URL = 'https://github.com/uliontse/mlgb'
DESCRIPTION = 'MLGB is a library that includes many models of CTR Prediction & Recommender System by TensorFlow & PyTorch.'
LONG_DESCRIPTION = pathlib.Path('README.md').read_text(encoding='utf-8')
VERSION = re.search(r'^__version__\s*=\s*[\'"]([^\'"]*)[\'"]', pathlib.Path('mlgb/__init__.py').read_text(), re.M).group(1)


setuptools.setup(
    name=NAME,
    version=VERSION,
    description=DESCRIPTION,
    long_description=LONG_DESCRIPTION,
    long_description_content_type='text/markdown',
    author=AUTHOR,
    author_email=AUTHOR_EMAIL,
    license='Apache-2.0',
    packages=setuptools.find_packages(),
    include_package_data=True,
    package_dir={'mlgb': 'mlgb'},
    url=HOMEPAGE_URL,
    project_urls={
        'Source': 'https://github.com/UlionTse/mlgb',
        'Changelog': 'https://github.com/UlionTse/mlgb/blob/main/change_log.txt',
        'Documentation': 'https://github.com/UlionTse/mlgb/blob/main/README.md',
    },
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: Apache Software License',
        'Programming Language :: Python :: 3 :: Only',
        # 'Programming Language :: Python :: 3.8',  # 2024.10
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3.10',
        'Programming Language :: Python :: 3.11',
        'Programming Language :: Python :: 3.12',
        'Topic :: Scientific/Engineering',
        'Topic :: Scientific/Engineering :: Artificial Intelligence',
        'Topic :: Software Development',
        'Topic :: Software Development :: Libraries',
        'Topic :: Software Development :: Libraries :: Python Modules',
    ],
    keywords=['Machine Learning', 'Deep Learning', 'CTR Prediction', 'Recommender System'],
    install_requires=[
        'numpy>=1.23.0',
        'pandas>=1.5.0',
        'scikit-learn>=1.3.0',
        'torch>=2.1.0',
        'tensorflow>=2.10.0',
        # 'tensorflow-addons>=0.21.0',
    ],
    python_requires='>=3.9',
    extras_require={'pypi': ['build>=0.10.0', 'twine>=4.0.2']},
    zip_safe=False,
)














